from copy import deepcopy
from logging import getLogger

from pyinaturalist.constants import API_V2, V2_OBS_ORDER_BY_PROPERTIES, JsonResponse, RequestParams
from pyinaturalist.converters import convert_all_coordinates, convert_all_timestamps
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.paginator import paginate_all
from pyinaturalist.request_params import validate_multiple_choice_param
from pyinaturalist.session import get, post

logger = getLogger(__name__)


# TODO: Send GET request with RISON if it fits in URL character limit?
# TODO: Code reuse, cleanup, etc.
@document_request_params(
    *docs._get_observations,
    docs._observation_v2,
    docs._pagination,
    docs._only_id,
    docs._access_token,
)
def get_observations(**params) -> JsonResponse:
    """Search observations

    .. rubric:: Notes

    * :fas:`lock-open` :ref:`Optional authentication <auth>` (For private/obscured coordinates)
    * API reference: :v2:`GET /observations <Observations/get_observations>`
    * See `iNaturalist API v2 documentation <https://api.inaturalist.org/v2/docs>`_ for details on
      selecting return fields using ``fields`` parameter
    * :fa:`exclamation-triangle` **Provisional:** This is for testing purposes only, and will change in future releases

    Examples:

        Get observations of Monarch butterflies with photos + public location info,
        on a specific date in the province of Saskatchewan, CA (place ID 7953),
        and return all available fields:

        >>> response = get_observations(
        >>>     taxon_name='Danaus plexippus',
        >>>     created_on='2020-08-27',
        >>>     photos=True,
        >>>     geo=True,
        >>>     geoprivacy='open',
        >>>     place_id=7953,
        >>>     fields='all',
        >>> )

        Get basic info for observations in response:

        >>> pprint(response)
        '[57754375] Species: Danaus plexippus (Monarch) observed by samroom on 2020-08-27 at Railway Ave, Wilcox, SK'
        '[57707611] Species: Danaus plexippus (Monarch) observed by ingridt3 on 2020-08-26 at Michener Dr, Regina, SK'

        Return all response fields *except* identifications:

        >>> response = get_observations(id=14150125, except_fields=['identifications'])

        Search for observations with a given observation field:

        >>> response = get_observations(observation_fields=['Species count'])

        Or observation field value:

        >>> response = get_observations(observation_fields={'Species count': 2})

        .. dropdown:: Example Response (default/minimal)
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observations_v2_minimal.py

        .. dropdown:: Example Response (all fields)
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_observations_v2_full.py

    Returns:
        Response dict containing observation records
    """
    params = validate_multiple_choice_param(params, 'order_by', V2_OBS_ORDER_BY_PROPERTIES)
    except_fields = params.pop('except_fields', None)

    if params.get('fields') and except_fields:
        raise ValueError('Cannot use both fields and except_fields')

    # Request all fields except those specified
    if except_fields:
        params['fields'] = deepcopy(ALL_OBS_FIELDS)
        for k in except_fields:
            params['fields'].pop(k, None)

    # If no field selections are specified (or all fields), use GET method
    if params.get('fields') in ['all', None]:
        observations = _get_observations(params)

    # If field selections are specified, use POST method and put fields in request body
    else:
        observations = _get_post_observations(params)

    observations['results'] = convert_all_coordinates(observations['results'])
    observations['results'] = convert_all_timestamps(observations['results'])
    return observations


def _get_observations(params: RequestParams) -> JsonResponse:
    if params.get('page') == 'all':
        return paginate_all(get, f'{API_V2}/observations', method='id', **params)
    else:
        return get(f'{API_V2}/observations', **params).json()


def _get_post_observations(params: RequestParams) -> JsonResponse:
    """Use POST method to get observations with selection of return fields.
    This allows for more complex requests that may exceed the max length of a GET URL.
    """
    headers = {'X-HTTP-Method-Override': 'GET'}
    fields = {'fields': params.pop('fields')}

    if params.get('page') == 'all':
        return paginate_all(
            post,
            f'{API_V2}/observations',
            method='id',
            headers=headers,
            json=fields,
            **params,
        )
    else:
        return post(
            f'{API_V2}/observations',
            headers=headers,
            json=fields,
            **params,
        ).json()


# The full `fields` value to request all observation details
ALL_OBS_FIELDS = {
    'annotations': {
        'controlled_attribute': {'id': True, 'label': True, 'multivalued': True},
        'controlled_value': {'id': True, 'label': True, 'multivalued': True},
        'user': {'login': True, 'icon_url': True},
        'vote_score': True,
        'votes': {'vote_flag': True, 'user': {'login': True, 'icon_url': True}},
    },
    'application': {'icon': True, 'name': True, 'url': True},
    'comments': {
        'body': True,
        'created_at': True,
        'flags': {'id': True},
        'hidden': True,
        'id': True,
        'moderator_actions': {
            'action': True,
            'id': True,
            'created_at': True,
            'reason': True,
            'user': {'login': True, 'icon_url': True},
        },
        'spam': True,
        'user': {'login': True, 'icon_url': True},
    },
    'community_taxon': {
        'ancestry': True,
        'ancestor_ids': True,
        'ancestors': {
            'id': True,
            'uuid': True,
            'name': True,
            'iconic_taxon_name': True,
            'is_active': True,
            'preferred_common_name': True,
            'rank': True,
            'rank_level': True,
        },
        'default_photo': {
            'attribution': True,
            'license_code': True,
            'url': True,
            'square_url': True,
        },
        'iconic_taxon_name': True,
        'id': True,
        'is_active': True,
        'name': True,
        'preferred_common_name': True,
        'rank': True,
        'rank_level': True,
    },
    'created_at': True,
    'description': True,
    'faves': {'id': True, 'user': {'login': True, 'icon_url': True}},
    'flags': {'id': True, 'flag': True, 'resolved': True},
    'geojson': True,
    'geoprivacy': True,
    'id': True,
    'identifications': {
        'body': True,
        'category': True,
        'created_at': True,
        'current': True,
        'disagreement': True,
        'flags': {'id': True},
        'hidden': True,
        'moderator_actions': {
            'action': True,
            'id': True,
            'created_at': True,
            'reason': True,
            'user': {'login': True, 'icon_url': True},
        },
        'previous_observation_taxon': {
            'ancestry': True,
            'ancestor_ids': True,
            'ancestors': {
                'id': True,
                'uuid': True,
                'name': True,
                'iconic_taxon_name': True,
                'is_active': True,
                'preferred_common_name': True,
                'rank': True,
                'rank_level': True,
            },
            'default_photo': {
                'attribution': True,
                'license_code': True,
                'url': True,
                'square_url': True,
            },
            'iconic_taxon_name': True,
            'id': True,
            'is_active': True,
            'name': True,
            'preferred_common_name': True,
            'rank': True,
            'rank_level': True,
        },
        'spam': True,
        'taxon': {
            'ancestry': True,
            'ancestor_ids': True,
            'ancestors': {
                'id': True,
                'uuid': True,
                'name': True,
                'iconic_taxon_name': True,
                'is_active': True,
                'preferred_common_name': True,
                'rank': True,
                'rank_level': True,
            },
            'default_photo': {
                'attribution': True,
                'license_code': True,
                'url': True,
                'square_url': True,
            },
            'iconic_taxon_name': True,
            'id': True,
            'is_active': True,
            'name': True,
            'preferred_common_name': True,
            'rank': True,
            'rank_level': True,
        },
        'taxon_change': {'id': True, 'type': True},
        'updated_at': True,
        'user': {'login': True, 'icon_url': True, 'id': True},
        'uuid': True,
        'vision': True,
    },
    'identifications_most_agree': True,
    'latitude': True,
    'license_code': True,
    'location': True,
    'longitude': True,
    'map_scale': True,
    'non_traditional_projects': {
        'current_user_is_member': True,
        'project_user': {'user': {'login': True, 'icon_url': True}},
        'project': {
            'admins': {'user_id': True},
            'icon': True,
            'project_observation_fields': {
                'id': True,
                'observation_field': {
                    'allowed_values': True,
                    'datatype': True,
                    'description': True,
                    'id': True,
                    'name': True,
                },
            },
            'slug': True,
            'title': True,
        },
    },
    'obscured': True,
    'observed_on': True,
    'observed_time_zone': True,
    'ofvs': {
        'observation_field': {
            'allowed_values': True,
            'datatype': True,
            'description': True,
            'name': True,
            'taxon': {'name': True},
            'uuid': True,
        },
        'user': {'login': True, 'icon_url': True},
        'uuid': True,
        'value': True,
        'taxon': {
            'ancestry': True,
            'ancestor_ids': True,
            'ancestors': {
                'id': True,
                'uuid': True,
                'name': True,
                'iconic_taxon_name': True,
                'is_active': True,
                'preferred_common_name': True,
                'rank': True,
                'rank_level': True,
            },
            'default_photo': {
                'attribution': True,
                'license_code': True,
                'url': True,
                'square_url': True,
            },
            'iconic_taxon_name': True,
            'id': True,
            'is_active': True,
            'name': True,
            'preferred_common_name': True,
            'rank': True,
            'rank_level': True,
        },
    },
    'outlinks': {'source': True, 'url': True},
    'photos': {
        'id': True,
        'uuid': True,
        'url': True,
        'license_code': True,
        'flags': {'id': True, 'flag': True, 'resolved': True},
    },
    'place_guess': True,
    'place_ids': True,
    'positional_accuracy': True,
    'preferences': {'prefers_community_taxon': True},
    'private_geojson': True,
    'private_place_guess': True,
    'private_place_ids': True,
    'project_observations': {
        'current_user_is_member': True,
        'preferences': {'allows_curator_coordinate_access': True},
        'project': {
            'admins': {'user_id': True},
            'icon': True,
            'project_observation_fields': {
                'id': True,
                'observation_field': {
                    'allowed_values': True,
                    'datatype': True,
                    'description': True,
                    'id': True,
                    'name': True,
                },
            },
            'slug': True,
            'title': True,
        },
        'uuid': True,
    },
    'public_positional_accuracy': True,
    'quality_grade': True,
    'reviewed_by': True,
    'sounds': {
        'file_url': True,
        'file_content_type': True,
        'id': True,
        'license_code': True,
        'play_local': True,
        'url': True,
        'uuid': True,
    },
    'tags': True,
    'taxon': {
        'ancestry': True,
        'ancestor_ids': True,
        'ancestors': {
            'id': True,
            'uuid': True,
            'name': True,
            'iconic_taxon_name': True,
            'is_active': True,
            'preferred_common_name': True,
            'rank': True,
            'rank_level': True,
        },
        'default_photo': {
            'attribution': True,
            'license_code': True,
            'url': True,
            'square_url': True,
        },
        'iconic_taxon_name': True,
        'id': True,
        'is_active': True,
        'name': True,
        'preferred_common_name': True,
        'rank': True,
        'rank_level': True,
    },
    'taxon_geoprivacy': True,
    'time_observed_at': True,
    'time_zone': True,
    'user': {
        'login': True,
        'icon_url': True,
        'id': True,
        'name': True,
        'observations_count': True,
        'preferences': {
            'prefers_community_taxa': True,
            'prefers_observation_fields_by': True,
            'prefers_project_addition_by': True,
        },
    },
    'viewer_trusted_by_observer': True,
    'votes': {
        'id': True,
        'user': {'login': True, 'icon_url': True, 'id': True},
        'vote_flag': True,
        'vote_scope': True,
    },
}



"""Models for taxon checklists (aka "original life lists")"""

from typing import List, Optional

from pyinaturalist.constants import ESTABLISTMENT_MEANS, DateTime, TableRow
from pyinaturalist.models import (
    BaseModel,
    LazyProperty,
    Place,
    User,
    datetime_field,
    define_model,
    define_model_custom_init,
    field,
)


@define_model
class EstablishmentMeans(BaseModel):
    """:fa:`exclamation-triangle` The establishment means for a taxon in a given location"""

    establishment_means: str = field(
        default=None,
        options=ESTABLISTMENT_MEANS,
        converter=lambda x: str(x).lower(),
        doc='Establishment means label',
    )
    establishment_means_description: str = field(
        default=None, options=ESTABLISTMENT_MEANS, doc='Establishment means description'
    )
    place: property = LazyProperty(
        Place.from_json, type=Place, doc='Location that the establishment means applies to'
    )

    # TODO: Is establishment_means_label ever different from establishment_means?
    @property
    def establishment_means_label(self) -> str:
        return self.establishment_means

    @establishment_means_label.setter
    def establishment_means_label(self, value: str):
        self.establishment_means = value

    def __str__(self) -> str:
        return f'EstablishmentMeans({self.establishment_means})'


@define_model
class Checklist(BaseModel):
    """:fa:`dove` :fa:`list` A taxon checklist (aka
    `"original life list" <https://www.inaturalist.org/blog/43337-upcoming-changes-to-lists>`_)
    """

    title: str = field(default=None)


@define_model_custom_init
class ListedTaxon(EstablishmentMeans):
    """:fa:`dove` :fa:`list` A taxon with additional stats associated with a list
    (aka `"original life list" <https://www.inaturalist.org/blog/43337-upcoming-changes-to-lists>`_),
    based on the schema of:

    * ``Taxon.listed_taxa`` from :v1:`GET /taxa/{id} <Taxa/get_taxa_id>`
    * ``TaxonSummary.listed_taxon`` from
      :v1:`GET /observations/{id}/taxon_summary <Observations/get_observations_id_taxon_summary>`
    """

    comments_count: int = field(default=0, doc='Number of comments for this listed taxon')
    created_at: DateTime = datetime_field(doc='Date and time the record was created')
    description: str = field(default=None, doc='Listed taxon description')
    first_observation_id: int = field(default=None, doc='Oldest recent observation ID in the list')
    last_observation_id: int = field(default=None, doc='Most recent observation ID in the list')
    manually_added: bool = field(
        default=None, doc='Indicates if the taxon was manually added to the list'
    )
    observations_count: int = field(
        default=0, doc='Number of observations of this taxon in the list'
    )
    occurrence_status_level: int = field(default=None, doc='')
    primary_listing: bool = field(
        default=None, doc='Indicates if this is the primary listing for this taxon'
    )
    source_id: int = field(default=None, doc='')
    taxon_id: int = field(default=None, doc='')
    taxon_range_id: int = field(default=None, doc='')
    updated_at: DateTime = datetime_field(doc='Date and time the record was last updated')

    list: Checklist = field(default=None, converter=Checklist.from_json, doc='Associated checklist')
    updater: User = field(
        default=None, converter=User.from_json, doc='User that last updated the record'
    )
    user: User = field(default=None, converter=User.from_json, doc='User that created the record')

    # Attributes that will only be used during init and then omitted
    temp_attrs = ['list_id', 'place_id', 'updater_id', 'user_id']

    def __init__(
        self,
        list_id: Optional[int] = None,
        place_id: Optional[int] = None,
        updater_id: Optional[int] = None,
        user_id: Optional[int] = None,
        **kwargs,
    ):
        self.__attrs_init__(**kwargs)  # type: ignore
        self.list = self.list or Checklist()
        if list_id:
            self.list_id = list_id
        if place_id:
            self.place_id = place_id
        if updater_id:
            self.updater_id = updater_id
        if user_id:
            self.user_id = user_id

    # Wrapper properties to handle some more inconsistencies between similar endpoints
    @property
    def list_id(self) -> Optional[int]:
        return self.list.id if self.list else None

    @list_id.setter
    def list_id(self, value: int):
        self.list.id = value

    @property
    def place_id(self) -> Optional[int]:
        return self.place.id if self.place else None

    @place_id.setter
    def place_id(self, value: int):
        self.place = self.place or Place()
        self.place.id = value

    @property
    def updater_id(self) -> Optional[int]:
        return self.updater.id if self.updater else None

    @updater_id.setter
    def updater_id(self, value: int):
        self.updater = self.updater or User()
        self.updater.id = value

    @property
    def user_id(self) -> Optional[int]:
        return self.user.id if self.user else None

    @user_id.setter
    def user_id(self, value: int):
        self.user = self.user or User()
        self.user.id = value

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Taxon ID': self.taxon_id,
            'Place ID': self.place.id,
            'Life list': self.list.title or self.list.id,
            'Establishment means': self.establishment_means,
            'Observations': self.observations_count,
            'Comments': self.comments_count,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'taxon_id', 'place', 'establishment_means', 'observations_count']

    def __str__(self) -> str:
        return BaseModel.__str__(self)



from typing import Optional

from attr import fields_dict

from pyinaturalist.constants import MultiInt
from pyinaturalist.controllers import BaseController
from pyinaturalist.converters import ensure_list
from pyinaturalist.docs import copy_doc_signature
from pyinaturalist.docs import templates as docs
from pyinaturalist.models import Taxon
from pyinaturalist.paginator import IDPaginator, Paginator
from pyinaturalist.v1 import get_taxa, get_taxa_autocomplete, get_taxa_by_id

IDS_PER_REQUEST = 30


class TaxonController(BaseController):
    """:fa:`dove` Controller for Taxon requests"""

    def __call__(self, taxon_id: int, **kwargs) -> Optional[Taxon]:
        """Get a single taxon by ID

        Example:
            >>> client.taxa(343248)

        Args:
            taxon_id: A single taxon ID
            locale: Locale preference for taxon common names
            preferred_place_id: Place preference for regional taxon common names
            all_names: Include all taxon names in the response
        """
        return self.from_ids(taxon_id, **kwargs).one()

    def from_ids(self, taxon_ids: MultiInt, **params) -> Paginator[Taxon]:
        """Get one or more taxa by ID

        .. rubric:: Notes

        * API reference: :v1:`GET /taxa/{id} <Taxa/get_taxa_id>`


        Example:
            >>> client.get_taxa_by_id([3, 343248])

        Args:
            taxon_ids: One or more taxon IDs
            locale: Locale preference for taxon common names
            preferred_place_id: Place preference for regional taxon common names
            all_names: Include all taxon names in the response
        """

        params = self.client.add_defaults(get_taxa_by_id, params)
        return IDPaginator(
            get_taxa_by_id,
            Taxon,
            ids=ensure_list(taxon_ids),
            ids_per_request=IDS_PER_REQUEST,
            **params,
        )

    @copy_doc_signature(docs._taxon_params)
    def autocomplete(self, **params) -> Paginator[Taxon]:
        """Given a query string, return taxa with names starting with the search term

        .. rubric:: Notes

        * API reference: :v1:`GET /taxa/autocomplete <Taxa/get_taxa_autocomplete>`
        * There appears to currently be a bug in the API that causes ``per_page`` to not have any effect.

        Example:
            >>> client.taxa.autocomplete(q='vespi')
        """
        return self.client.paginate(get_taxa_autocomplete, Taxon, **params)

    @copy_doc_signature(docs._taxon_params, docs._taxon_id_params)
    def search(self, **params) -> Paginator[Taxon]:
        """Search taxa

        .. rubric:: Notes

        * API reference: :v1:`GET /taxa <Taxa/get_taxa>`

        Example:
            >>> client.taxa.search(q='vespi', rank=['genus', 'family'])
        """
        return self.client.paginate(get_taxa, Taxon, **params)

    def populate(self, taxon: Taxon, **params) -> Taxon:
        """Update a partial Taxon record with full taxonomy info, including ancestors + children

        Args:
            taxon: A partial Taxon record

        Returns:
            The same Taxon record, updated with full taxonomy info
        """
        # Don't overwrite these keys if set by a previous API call
        preserve_keys = {'listed_taxa', 'matched_term', 'names'}

        full_taxon = self.from_ids(taxon.id, **params).one()
        for key in set(fields_dict(Taxon).keys()) - preserve_keys:
            # Use getters/setters for LazyProperty instead of temp attrs (cls.foo vs cls._foo)
            if hasattr(taxon, key.lstrip('_')):
                key = key.lstrip('_')
            setattr(taxon, key, getattr(full_taxon, key))
        return taxon



# TODO: "optional auth" option
# TODO: Improve Sphinx docs generated for controller attributes
# TODO: Use a custom template or directive to generate summary of all controller methods
from asyncio import AbstractEventLoop
from inspect import ismethod
from logging import getLogger
from typing import Any, Callable, Dict, Optional, Type

from pyinaturalist.auth import get_access_token
from pyinaturalist.constants import RequestParams
from pyinaturalist.controllers import (
    AnnotationController,
    ObservationController,
    PlaceController,
    ProjectController,
    SearchController,
    TaxonController,
    UserController,
)
from pyinaturalist.models import T
from pyinaturalist.paginator import Paginator
from pyinaturalist.request_params import get_valid_kwargs, strip_empty_values
from pyinaturalist.session import ClientSession

logger = getLogger(__name__)


# 'iNatClient' is nonstandard casing, but 'InatClient' just looks wrong. Deal with it, pep8.
class iNatClient:
    """API client class that provides an object-oriented interface to the iNaturalist API.

    **WIP/Experimental**

    See:

    * Usage guide: :ref:`api-client`
    * Query functions: :ref:`Controller classes <controllers>`

    Controllers:

    * :fa:`tag` :py:class:`annotations <.AnnotationController>`
    * :fa:`binoculars` :py:class:`observations <.ObservationController>`
    * :fa:`location-dot` :py:class:`places <.PlaceController>`
    * :fa:`users` :py:class:`projects <.ProjectController>`
    * :fa:`search` :py:class:`search <.SearchController>`
    * :fa:`dove` :py:class:`taxa <.TaxonController>`
    * :fa:`user` :py:class:`users <.UserController>`

    Args:
        creds: Optional arguments for :py:func:`.get_access_token`, used to get and refresh access
            tokens as needed. Using a keyring instead is recommended, though.
        default_params: Default request parameters to pass to any applicable API requests
        dry_run: Just log all requests instead of sending real requests
        loop: An event loop to run any executors used for async iteration
        session: Session object to use instead of creating a new one
        kwargs: Keyword arguments for :py:class:`.ClientSession`
    """

    def __init__(
        self,
        creds: Optional[Dict[str, str]] = None,
        default_params: Optional[Dict[str, Any]] = None,
        dry_run: bool = False,
        loop: Optional[AbstractEventLoop] = None,
        session: Optional[ClientSession] = None,
        **kwargs,
    ):
        self.creds = creds or {}
        self.default_params = default_params or {}
        self.dry_run = dry_run
        self.loop = loop
        self.session = session or ClientSession(**kwargs)

        self._access_token = None
        self._token_expires = None

        # Controllers
        self.annotations = AnnotationController(
            self
        )  #: Interface for :py:class:`annotation requests <.AnnotationController>`
        self.observations = ObservationController(
            self
        )  #: Interface for :py:class:`observation requests <.ObservationController>`
        self.places = PlaceController(
            self
        )  #: Interface for :py:class:`place requests <.PlaceController>`
        self.projects = ProjectController(
            self
        )  #: Interface for :py:class:`project requests <.ProjectController>`
        self.search = SearchController(
            self
        )  #: Unified :py:meth:`text search <.SearchController.__call__>`
        self.taxa = TaxonController(
            self
        )  #: Interface for :py:class:`taxon requests <.TaxonController>`
        self.users = UserController(
            self
        )  #: Interface for :py:class:`user requests <.UserController>`

    def add_defaults(
        self, request_function, kwargs: Optional[RequestParams] = None, auth: bool = False
    ) -> RequestParams:
        """Add any applicable client settings to request parameters before sending a request.
        Explicit keyword arguments will override any client settings.
        """
        kwargs = strip_empty_values(kwargs or {})
        client_kwargs = {'dry_run': self.dry_run, 'session': self.session}

        # Add access token and default params, if applicable
        if auth:
            client_kwargs['access_token'] = get_access_token(**self.creds)  # type: ignore
        client_kwargs.update(get_valid_kwargs(request_function, self.default_params))
        for k, v in client_kwargs.items():
            kwargs.setdefault(k, v)

        # If we're directly calling a ClientSession method, we don't need to pass a session object
        if ismethod(request_function):
            kwargs.pop('session')

        return kwargs

    def paginate(
        self,
        request_function: Callable,
        model: Type[T],
        auth: bool = False,
        cls: Type[Paginator] = Paginator,
        **kwargs,
    ) -> Paginator[T]:
        """Create a paginator for a request, with client settings applied

        Args:
            request_function: The API request function to call
            model: Model class used for the response
            auth: Indicates that the request requires authentication
            cls: Alternative Paginator class to use
            params: Original request parameters
        """
        kwargs = self.add_defaults(request_function, kwargs, auth)
        return cls(request_function, model, loop=self.loop, **kwargs)

    def request(self, request_function: Callable, *args, auth: bool = False, **kwargs):
        """Send a request, with client settings applied.

        Args:
            request_function: The API request function to call
            auth: Indicates that the request requires authentication
            params: Original request parameters

        Returns:
            Results of ``request_function()``
        """
        kwargs = self.add_defaults(request_function, kwargs, auth)
        return request_function(*args, **kwargs)



"""
Functions to access the iNaturalist API v1
See: http://api.inaturalist.org/v1/docs/

Most recent API version tested: 1.3.0
"""

# ruff: noqa: F401, F403
from pyinaturalist.v1.controlled_terms import get_controlled_terms, get_controlled_terms_for_taxon
from pyinaturalist.v1.identifications import get_identifications, get_identifications_by_id
from pyinaturalist.v1.messages import get_message_by_id, get_messages, get_unread_message_count
from pyinaturalist.v1.observation_fields import delete_observation_field, set_observation_field
from pyinaturalist.v1.observations import (
    create_observation,
    delete_observation,
    get_observation,
    get_observation_histogram,
    get_observation_identifiers,
    get_observation_observers,
    get_observation_popular_field_values,
    get_observation_species_counts,
    get_observation_taxon_summary,
    get_observation_taxonomy,
    get_observations,
    get_observations_by_id,
    update_observation,
    upload,
)
from pyinaturalist.v1.places import get_places_autocomplete, get_places_by_id, get_places_nearby
from pyinaturalist.v1.posts import get_posts
from pyinaturalist.v1.projects import (
    add_project_observation,
    add_project_users,
    delete_project_observation,
    delete_project_users,
    get_projects,
    get_projects_by_id,
    update_project,
)
from pyinaturalist.v1.search import search
from pyinaturalist.v1.taxa import (
    get_life_list_metadata,
    get_taxa,
    get_taxa_autocomplete,
    get_taxa_by_id,
    get_taxa_map_layers,
)
from pyinaturalist.v1.users import get_current_user, get_user_by_id, get_users_autocomplete



from logging import getLogger
from typing import Optional

from pyinaturalist.constants import (
    API_V1,
    PROJECT_ORDER_BY_PROPERTIES,
    IntOrStr,
    JsonResponse,
    MultiInt,
    MultiIntOrStr,
)
from pyinaturalist.converters import convert_all_coordinates, convert_all_timestamps, ensure_list
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.paginator import paginate_all
from pyinaturalist.request_params import split_common_params, validate_multiple_choice_param
from pyinaturalist.session import ClientSession, delete, get, get_local_session, post, put

logger = getLogger(__name__)


@document_request_params(docs._projects_params, docs._pagination)
def get_projects(**params) -> JsonResponse:
    """Search projects

    .. rubric:: Notes

    * API reference: :v1:`GET /projects <Projects/get_projects>`

    Example:
        Search for projects about invasive species within 400km of Vancouver, BC:

        >>> response = get_projects(
        >>>     q='invasive',
        >>>     lat=49.27,
        >>>     lng=-123.08,
        >>>     radius=400,
        >>>     order_by='distance',
        >>> )

        Show basic info for projects in response:

        >>> pprint(response)
        [8291    ] PNW Invasive Plant EDDR
        [19200   ] King County (WA) Noxious and Invasive Weeds
        [102925  ] Keechelus/Kachess Invasive Plants
        ...

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_projects.py

    Returns:
        Response dict containing project records
    """
    validate_multiple_choice_param(params, 'order_by', PROJECT_ORDER_BY_PROPERTIES)
    if params.get('page') == 'all':
        projects = paginate_all(get, f'{API_V1}/projects', **params)
    else:
        projects = get(f'{API_V1}/projects', **params).json()

    projects['results'] = convert_all_coordinates(projects['results'])
    projects['results'] = convert_all_timestamps(projects['results'])
    return projects


def get_projects_by_id(
    project_id: MultiIntOrStr,
    rule_details: Optional[bool] = None,
    force_refresh: bool = False,
    session: Optional[ClientSession] = None,
    **params,
) -> JsonResponse:
    """Get one or more projects by ID

    .. rubric:: Notes

    * API reference: :v1:`GET /projects/{id} <Projects/get_projects_id>`

    Example:
        >>> response = get_projects_by_id([8348, 6432])
        >>> pprint(response)
        [8348] Tucson High Native and Invasive Species Inventory
        [6432] CBWN Invasive Plants

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/get_projects_by_id.py

    Args:
        project_id: Get projects with this ID. Multiple values are allowed.
        rule_details: Return more information about project rules, for example return a full taxon
            object instead of simply an ID
        force_refresh: Force a refresh of the project record from the API, bypassing both the local
            and CDN caches

    Returns:
        Response dict containing project records
    """
    # If needed, invalidate CDN cache to ensure a fresh response
    session = session or get_local_session()
    if force_refresh:
        params.update(session.get_refresh_params(f'/projects/{project_id}'))

    params['rule_details'] = rule_details
    response = session.request(
        'GET',
        f'{API_V1}/projects',
        ids=project_id,  # type: ignore[arg-type]
        allow_str_ids=True,
        **params,
    )

    projects = response.json()
    projects['results'] = convert_all_coordinates(projects['results'])
    projects['results'] = convert_all_timestamps(projects['results'])
    return projects


@document_request_params(docs._project_observation_params)
def add_project_observation(
    project_id: int, observation_id: int, access_token: Optional[str] = None, **params
) -> JsonResponse:
    """Add an observation to a project

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`POST projects/{id}/add <Projects/post_projects_id_add>`
    * API reference: :v1:`POST /project_observations <Project_Observations/post_project_observations>`

    Example:
        >>> add_project_observation(24237, 1234, access_token)

        .. dropdown:: Example Response
            :color: primary
            :icon: code-square

            .. literalinclude:: ../sample_data/add_project_observation.json
                :language: JSON

    Returns:
        Information about the added project observation
    """
    response = post(
        f'{API_V1}/project_observations',
        access_token=access_token,
        json={'observation_id': observation_id, 'project_id': project_id},
        **params,
    )
    return response.json()


def add_project_users(project_id: IntOrStr, user_ids: MultiInt, **params) -> JsonResponse:
    """Add users to project observation rules

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * This only affects observation rules, **not** project membership

    Args:
        project_id: Either numeric project ID or URL slug
        user_ids: One or more user IDs to add. Only accepts numeric IDs.

    Returns:
        The updated project record
    """
    project = get_projects_by_id(project_id, force_refresh=True)['results'][0]
    existing_rules = project.get('project_observation_rules', [])
    existing_user_ids = [
        rule['operand_id'] for rule in existing_rules if rule['operand_type'] == 'User'
    ]
    new_rules = []

    # Minimize payload by sending only new rules
    for user_id in ensure_list(user_ids):
        if user_id not in existing_user_ids:
            new_rules.append(
                {'operand_id': user_id, 'operand_type': 'User', 'operator': 'observed_by_user?'}
            )
        else:
            logger.warning(f'User {user_id} is already in project rules for {project_id}')

    if not new_rules:
        logger.info('No project rules to update')
        return project

    return update_project(project_id, project_observation_rules_attributes=new_rules, **params)


# TODO: This may not yet be working as intended
@document_request_params(docs._project_observation_params)
def delete_project_observation(
    project_id: int, observation_id: int, access_token: Optional[str] = None, **params
):
    """Remove an observation from a project

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * API reference: :v1:`DELETE /projects/{id}/remove <Projects/delete_projects_id_remove>`
    * API reference: :v1:`DELETE /project_observations <Project_Observations/delete_project_observations_id>`

    Example:

        >>> delete_project_observation(24237, 1234, access_token=access_token)
    """
    return delete(
        f'{API_V1}/projects/{project_id}/remove',
        access_token=access_token,
        json={'observation_id': observation_id},
        **params,
    )

    # This version takes a separate 'project observation' ID (from association table?)
    # delete_v1(
    #     'project_observations',
    #     access_token=access_token,
    #     json={'observation_id': observation_id, 'project_id': project_id},
    #     **params,
    # )


def delete_project_users(project_id: IntOrStr, user_ids: MultiInt, **params) -> JsonResponse:
    """Remove users from project observation rules

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * This only affects observation rules, **not** project membership

    Args:
        project_id: Either numeric project ID or URL slug
        user_ids: One or more user IDs to remove. Only accepts numeric IDs.

    Returns:
        The updated project record
    """
    # Get and modify existing rules

    project = get_projects_by_id(project_id, force_refresh=True)['results'][0]
    existing_rules = project.get('project_observation_rules', [])
    delete_rules = []
    str_user_ids = [str(user_id) for user_id in ensure_list(user_ids)]

    for rule in existing_rules:
        if rule['operand_type'] == 'User' and str(rule['operand_id']) in str_user_ids:
            rule['_destroy'] = True
            delete_rules.append(rule)

    if not delete_rules:
        logger.info('No project rules to update')
        return project

    # Update project and validate results
    project = update_project(
        project_id, project_observation_rules_attributes=delete_rules, **params
    )
    _validate_removed_users(project, user_ids)
    return project


@document_request_params(docs._project_update_params)
def update_project(project_id: IntOrStr, **params) -> JsonResponse:
    """Update a project

    .. rubric:: Notes

    * :fa:`lock` :ref:`Requires authentication <auth>`
    * Undocumented endpoint; may be subject to braking changes in the future
    * ``admin_attributes`` and ``project_observation_rules_attributes`` each accept a list of dicts
      in the formats shown below. These can be obtained from :py:func:`get_projects`, modified, and
      then passed to this function::

        {
            "admin_attributes": [
                {"id": int, "role": str, "user_id": int, "_destroy": bool},
            ],
            "project_observation_rules_attributes": [
                {"operator": str, "operand_type": str, "operand_id": int, "id": int, "_destroy": bool},
            ],
        }

    Example:

        >>> update_project(
        ...     'api-test-project',
        ...     title='Test Project',
        ...     description='This is a test project',
        ...     prefers_rule_native=True,
        ...     access_token=access_token,
        ... )

    Returns:
        The updated project record
    """
    # Split API request params from common function args
    params, kwargs = split_common_params(params)
    kwargs['timeout'] = kwargs.get('timeout') or 60  # This endpoint can be a bit slow

    response = put(
        f'{API_V1}/projects/{project_id}',
        json={'project': params},
        **kwargs,
    )
    return response.json()


def _validate_removed_users(project: JsonResponse, user_ids: MultiInt):
    """Validate that users have been removed from project observation rules"""
    updated_user_ids = [
        rule['operand_id']
        for rule in project.get('project_observation_rules', [])
        if rule['operand_type'] == 'User'
    ]

    failed_ids = set(updated_user_ids) & set(ensure_list(user_ids))
    if failed_ids:
        logger.warning(f'Failed to remove users {list(failed_ids)} from project {project["id"]}')



"""Utilities for formatting API responses and model objects, for convenience/readability when exploring data.
Not used directly by API functions.

These functions will accept any of the following:

* A JSON response
* A list of response objects
* A single response object
"""

import json
from datetime import date, datetime, timedelta
from logging import basicConfig, getLogger
from typing import List, Mapping, Type, Union

from requests import PreparedRequest, Response
from rich import pretty, print
from rich.box import SIMPLE_HEAVY
from rich.logging import RichHandler
from rich.table import Column, Table
from rich.tree import Tree

from pyinaturalist.constants import DATETIME_SHORT_FORMAT, ResponseResult
from pyinaturalist.converters import ensure_list
from pyinaturalist.models import (
    Annotation,
    BaseModel,
    BaseModelCollection,
    Comment,
    ControlledTerm,
    ControlledTermCount,
    ControlledTermValue,
    Flag,
    Identification,
    LifeList,
    ListedTaxon,
    Message,
    Observation,
    ObservationField,
    ObservationFieldValue,
    Photo,
    Place,
    Project,
    QualityMetric,
    ResponseOrObjects,
    SearchResult,
    Sound,
    Taxon,
    TaxonCount,
    TaxonSummary,
    User,
    Vote,
)
from pyinaturalist.models.taxon import make_tree
from pyinaturalist.paginator import Paginator

pretty.install()


# Default colors for table headers
HEADER_COLORS = {
    'Category': 'violet',
    'Comment': 'green',
    'Comments': 'blue',
    'Common name': 'blue',
    'Count': 'blue',
    'Created at': 'blue',
    'Date': 'blue',
    'Description': 'green',
    'Dimensions': 'blue',
    'Display name': 'violet',
    'Establishment means': 'violet',
    'From': 'magenta',
    'From CV': 'white',
    'Identifications': 'blue',
    'ID': 'cyan',
    'Label': 'green',
    'Latitude': 'blue',
    'License': 'green',
    'Life list': 'green',
    'Location': 'white',
    'Longitude': 'blue',
    'Name': 'magenta',
    'Observations': 'blue',
    'Observed on': 'blue',
    'Place ID': 'cyan',
    'Rank': 'violet',
    'Scientific name': 'green',
    'Score': 'green',
    'Subject': 'green',
    'Taxon ID': 'cyan',
    'Taxon': 'green',
    'Title': 'green',
    'To': 'magenta',
    'Type': 'blue',
    'URL': 'white',
    'User': 'magenta',
    'Username': 'magenta',
    'Value': 'green',
    'Votes': 'blue',
}

# Unique response attributes used to auto-detect response types
UNIQUE_RESPONSE_ATTRS = {
    'vote_score': Annotation,
    'disagreement': Identification,
    'thread_id': Message,
    'body': Comment,  # Subset of ID attrs; if it's not an ID, assume it's a comment
    'multivalued': ControlledTerm,
    'controlled_value': ControlledTermCount,
    'blocking': ControlledTermValue,
    'count_without_taxon': LifeList,
    'captive': Observation,
    'allowed_values': ObservationField,
    'field_id': ObservationFieldValue,
    'flag': Flag,
    'original_dimensions': Photo,
    'place_type': Place,
    'standard': Place,
    'project_type': Project,
    'score': SearchResult,
    'last_observation_id': ListedTaxon,
    'listed_taxon': TaxonSummary,
    'metric': QualityMetric,
    'native_sound_id': Sound,
    'count': TaxonCount,
    'rank': Taxon,
    'roles': User,
    'vote_flag': Vote,
}


def enable_logging(level: str = 'INFO', external_level: str = 'WARNING'):
    """Configure logging to standard output with prettier tracebacks, formatting, and terminal
    colors (if supported).

    If you prefer, logging can be configured with the stdlib ``logging`` module instead; this just
    provides some convenient defaults.

    Args:
        level: Logging level to use for pyinaturalist
        external_level: Logging level to use for other libraries
    """

    basicConfig(
        format='%(message)s',
        datefmt='[%m-%d %H:%M:%S]',
        handlers=[RichHandler(rich_tracebacks=True, markup=True)],
        level=external_level,
    )
    getLogger('pyinaturalist').setLevel(level)
    getLogger('pyinaturalist_convert').setLevel(level)


def pprint(values: ResponseOrObjects):
    """Pretty-print any model object or list into a condensed summary.

    **Experimental:** May also be used on most raw JSON API responses
    """
    try:
        print(format_table(values))
    except ValueError:
        print(values)


def pprint_tree(taxa: Union[Taxon, LifeList, List[Taxon]], **kwargs):
    """Pretty-print a taxon and its descendants as a tree.

    Args:
        taxa: A single taxon with children, a :py:class:`.LifeList`, or a list of taxa
    """
    if isinstance(taxa, LifeList) or (
        isinstance(taxa, list) and len(taxa) > 0 and isinstance(taxa[0], Taxon)
    ):
        taxa = make_tree(taxa)
    elif not isinstance(taxa, Taxon):
        raise ValueError('Invalid taxon or taxon collection')

    print(format_tree(taxa, **kwargs))


def detect_type(value: ResponseResult) -> Type[BaseModel]:
    """Attempt to determine the model class corresponding to an API result"""
    for key, cls in UNIQUE_RESPONSE_ATTRS.items():
        if key in value:
            return cls

    raise ValueError(f'Could not detect response type: {value}')


def ensure_model_list(values: ResponseOrObjects) -> List[BaseModel]:
    """If the given values are raw JSON responses, attempt to detect their type and convert to
    model objects
    """
    if isinstance(values, Paginator):
        return values.all()

    values = ensure_list(values)
    if isinstance(values, BaseModelCollection) or isinstance(values[0], BaseModel):
        return values  # type: ignore

    cls = detect_type(values[0])
    return [cls.from_json(value) for value in values]


def format_table(values: ResponseOrObjects) -> Union[Table, str]:
    """Format model objects as a table.

    If the model doesn't have a table format defined, return a basic list of stringified objects.
    """
    try:
        if isinstance(values, Table):
            return values

        values = ensure_model_list(values)
        headers = {k: HEADER_COLORS.get(k, '') for k in values[0]._row.keys()}
    except NotImplementedError:
        return '\n'.join([str(obj) for obj in ensure_model_list(values)])

    # Display any date/datetime values in short format
    def _str(value):
        if isinstance(value, (date, datetime)):
            return value.strftime(DATETIME_SHORT_FORMAT)
        return str(value) if value is not None else ''

    columns = [Column(header, style=style) for header, style in headers.items()]
    table = Table(*columns, box=SIMPLE_HEAVY, header_style='bold white', row_styles=['dim', 'none'])

    for obj in values:
        table.add_row(*[_str(value) for value in obj._row.values()])
    return table


def format_request(request: PreparedRequest, dry_run: bool = False) -> str:
    """Format HTTP request info"""
    headers = _format_headers(request.headers)
    body = _format_body(request.body)
    dry_run_str = ' (DRY RUN) ' if dry_run else ''
    return f'Request:{dry_run_str}\n{request.method} {request.url}\n{headers}\n{body}'


def format_response(response: Response) -> str:
    """Format HTTP response info, including whether it came from the cache"""
    headers = _format_headers(response.headers)
    error_msg = f' {response.text}' if not response.ok else ''

    def _expires_str():
        if response.expires:
            expires_delta = timedelta(seconds=response.expires_delta)
            return f'expires in {expires_delta}'
        else:
            return 'never expires'

    cached = f'cached; {_expires_str()}' if getattr(response, 'from_cache', False) else 'not cached'
    return f'Response ({cached}):\n{response.status_code} {response.reason}{error_msg}\n{headers}'


def format_tree(taxon: Taxon, **kwargs) -> Tree:
    """Format a taxon and its descendants as a tree"""
    node = Tree(f'{taxon.rich_full_name}', **kwargs)
    for child in taxon.children:
        node.add(format_tree(child))
    return node


def _format_headers(headers: Mapping[str, str]) -> str:
    ignore_headers = [
        'Access-Control-Allow-Headers',
        'Access-Control-Allow-Methods',
        'Access-Control-Allow-Origin',
        'X-Content-Type-Options',
    ]
    headers_dict = {k: v for k, v in headers.items() if k not in ignore_headers}
    if 'Authorization' in headers_dict:
        headers_dict['Authorization'] = '[REDACTED]'

    return '\n'.join([f'{k}: {v}' for k, v in headers_dict.items()])


def _format_body(body):
    if not body:
        return ''
    try:
        body = json.loads(body)
        for key in ['password', 'client_secret']:
            if key in body:
                body[key] = '[REDACTED]'
        return body
    except Exception:
        return '(non-JSON request body)'



"""Base class and utilities for data models"""

from collections import UserList
from copy import deepcopy
from datetime import datetime
from logging import getLogger
from os.path import expanduser
from pathlib import Path
from typing import Dict, Generic, List, Optional, Type, TypeVar

from attr import Factory, asdict, define, field, fields_dict

from pyinaturalist.constants import (
    DATETIME_SHORT_FORMAT,
    AnyFile,
    JsonResponse,
    ResponseOrFile,
    ResponseOrResults,
    TableRow,
)
from pyinaturalist.converters import ensure_list, try_int

# Optionally use ultrajson instead of stdlib json, if available
try:
    import ujson as json
except ImportError:
    import json  # type: ignore

T = TypeVar('T', bound='BaseModel')
TC = TypeVar('TC', bound='BaseModelCollection')
logger = getLogger(__name__)


@define(auto_attribs=False)
class BaseModel:
    """Base class for data models"""

    id: int = field(default=None, converter=try_int, metadata={'doc': 'Unique record ID'})
    uuid: str = field(default=None, metadata={'doc': 'Unversally unique record ID'})
    __is_nested: bool = field(default=False, repr=False, init=False)
    temp_attrs: List[str] = []

    @classmethod
    def copy(cls, obj: 'BaseModel') -> 'BaseModel':
        """Copy a model object. This is defined as a classmethod to easily initialize a subclass
        from a parent class instance. For copying an instance to the same type,
        :py:func:`copy.deepcopy` can be used.
        """
        kwargs = obj.to_dict(recurse=False)
        return cls(**kwargs)

    @classmethod
    def from_json(cls: Type[T], value: JsonResponse, **kwargs) -> T:
        """Initialize a single model object from an API response or response result.

        Omits any invalid fields and ``None`` values, so default factories are used instead
        (e.g. for empty dicts and lists).
        """
        value = value or {}
        if isinstance(value, cls):
            return value

        cls_attrs = {k.lstrip('_'): v for k, v in fields_dict(cls).items()}

        def is_valid_attr(k):
            return (k in cls_attrs and cls_attrs[k].init is True) or k in cls.temp_attrs

        valid_json = {k: v for k, v in value.items() if is_valid_attr(k) and v is not None}
        return cls(**valid_json, **kwargs)

    @classmethod
    def from_json_file(cls: Type[T], value: AnyFile) -> List[T]:
        """Initialize a collection of model objects from a JSON string, file path, or file-like object"""
        return cls.from_json_list(load_json(value))

    @classmethod
    def from_json_list(cls: Type[T], value: ResponseOrResults, **kwargs) -> List[T]:
        """Initialize a collection of model objects from an API response or response results"""
        return [cls.from_json(item, **kwargs) for item in ensure_list(value)]

    @property
    def _is_nested(self) -> bool:
        """Check if this object is nested within another object, and reset flag. Used for condensed
        output of nested model objects. This flag is temporarily set and later checked/reset by
        ``BaseModel.__rich_repr__``

        This is a somewhat messy compromise to allow all of the following to work:

        * Print only condensed output when printed as a nested object
        * Print full output when not printed as a nested object
        * Do this while working within rich's rendering method (for pretty syntax highlighting)
        """
        val = self.__is_nested
        self.__is_nested = False
        return val

    @property
    def _row(self) -> TableRow:
        """Get values and headers to display as a row in a table"""
        raise NotImplementedError

    @property
    def _str_attrs(self) -> List[str]:
        """Get the subset of attribute names to show in the model's string representation"""
        return [getattr(a, 'name', '') for a in self.__attrs_attrs__]

    def to_dict(self, keys: Optional[List[str]] = None, recurse: bool = True) -> JsonResponse:
        """Convert this object back to dict format

        Args:
            keys: Only keep the specified keys (attribute names)
            recurse: Recurse into nested model objects
        """

        def recursive_serializer(_inst, _attribute, value):
            if keys and _attribute and _attribute.name.lstrip('_') not in keys:
                return None
            return value.to_dict() if isinstance(value, BaseModel) else value

        obj_dict = asdict(
            self,
            filter=lambda a, _: a.init is True,
            retain_collection_types=True,
            recurse=recurse,
            value_serializer=recursive_serializer if recurse else None,
        )

        obj_dict = {k.lstrip('_'): v for k, v in obj_dict.items()}
        if keys:
            obj_dict = {k: v for k, v in obj_dict.items() if k in keys}
        return obj_dict

    def __rich_repr__(self):
        """Custom output to use when pretty-printed with rich. Compared to default rich behavior for
        attrs classes, this does the following:

        * Inform rich about all default values, including factories, so they will be excluded from
          output
        * Stringify datetime objects
        * Add lazy-loaded attributes to the output
        * Use condensed (str) representations of lazy-loaded attributes
        * Does not currently handle positional-only args (since we don't currently have any)

        This makes output much more readable, particularly within Jupyter or a REPL. Otherwise, some
        objects (especially observations) can turn into a huge wall of text several pages long.
        """
        from pyinaturalist.models import get_lazy_properties

        # If this is a nested object, show a condensed representation
        if self._is_nested:
            for a in self._str_attrs:
                value = getattr(self, a)
                if isinstance(value, datetime):
                    value = value.strftime(DATETIME_SHORT_FORMAT)
                yield a, value, None
            return

        # Handle regular public attributes
        public_attrs = [
            a for a in self.__attrs_attrs__ if a.repr is True and not a.name.startswith('_')
        ]
        for a in public_attrs:
            default = a.default.factory() if isinstance(a.default, Factory) else a.default
            value = getattr(self, a.name)
            value = str(value) if isinstance(value, datetime) else value
            yield a.name, value, default

        # Set temporary flag for condensed output (see _is_nested property for details)
        def set_nested(prop_value):
            if isinstance(prop_value, list):
                for v in prop_value:
                    v.__is_nested = True
            elif prop_value:
                prop_value.__is_nested = True

        # Handle LazyProperties
        for name, prop in get_lazy_properties(type(self)).items():
            prop_value = prop.__get__(self, type(self))
            set_nested(prop_value)
            yield name, prop_value

    def __str__(self):
        """Make a string representation based on a minimal set of attributes defined on each model,
        via ``_str_attrs``.
        """
        repr_attrs = {a: getattr(self, a) for a in self._str_attrs}
        for k, v in repr_attrs.items():
            if isinstance(v, datetime):
                repr_attrs[k] = v.strftime(DATETIME_SHORT_FORMAT)
        repr_attrs_str = ', '.join([f'{k}={v}' for k, v in repr_attrs.items()])
        return f'{self.__class__.__name__}({repr_attrs_str})'


@define(auto_attribs=False, order=False, slots=False)
class BaseModelCollection(BaseModel, UserList, Generic[T]):
    """Base class for data model collections. These will behave the same as lists but enable some
    additional operations on contained items.
    """

    data: List[T] = field(factory=list, init=False, repr=False)
    _id_map: Dict[int, T] = field(default=None, init=False, repr=False)

    @classmethod
    def copy(cls, obj):
        """Copy a model collection"""
        return cls(data=[deepcopy(item) for item in obj])

    @classmethod
    def from_json(cls: Type[TC], value: JsonResponse, **kwargs) -> TC:
        if 'results' in value:
            value = value['results']
        if 'data' not in value:
            value = {'data': value}
        return super(BaseModelCollection, cls).from_json(value)

    @classmethod
    def from_json_list(cls: Type[TC], value: JsonResponse, **kwargs) -> TC:  # type: ignore
        """For model collections, initializing from a list should return an instance of ``cls``
        instead of a builtin ``list``
        """
        return cls.from_json(value)

    @property
    def id_map(self) -> Dict[int, T]:
        """A mapping of objects by unique identifier"""
        if self._id_map is None:
            self._id_map = {obj.id: obj for obj in self.data}
        return self._id_map

    def deduplicate(self):
        """Remove any duplicates from this collection based on ID"""
        self.data = list(self.id_map.values())

    def get_count(self, id: int, count_field: str = 'count') -> int:
        """Get a count associated with the given ID.
        Returns 0 if the collection type is not countable or the ID doesn't exist.
        """
        return getattr(self.id_map.get(id), count_field, 0)

    def __str__(self) -> str:
        return '\n'.join([str(obj) for obj in self.data])


def load_json(value: ResponseOrFile) -> ResponseOrResults:
    """Load a JSON string, file path, or file-like object"""
    if not value:
        return {}
    if isinstance(value, (dict, list)):
        json_value = value
    elif isinstance(value, (Path, str)):
        with open(expanduser(str(value))) as f:
            json_value = json.load(f)
    else:
        json_value = json.load(value)

    if 'results' in json_value:
        json_value = json_value['results']
    return json_value



from typing import Optional

from pyinaturalist.constants import IntOrStr, ListResponse, MultiInt, MultiIntOrStr
from pyinaturalist.controllers import BaseController
from pyinaturalist.converters import ensure_list
from pyinaturalist.docs import copy_doc_signature
from pyinaturalist.docs import templates as docs
from pyinaturalist.models import Project
from pyinaturalist.paginator import Paginator
from pyinaturalist.v1 import (
    add_project_observation,
    add_project_users,
    delete_project_users,
    get_projects,
    get_projects_by_id,
    update_project,
)


class ProjectController(BaseController):
    """:fa:`users` Controller for Project requests"""

    def __call__(self, project_id: int, **kwargs) -> Optional[Project]:
        """Get a single project by ID

        Example:
            >>> client.projects(1234)

        Args:
            project_id: A single project ID
        """
        return self.from_ids(project_id, **kwargs).one()

    def from_ids(self, project_ids: MultiIntOrStr, **params) -> Paginator[Project]:
        """Get projects by ID

        Example:
            >>> client.projects.from_id([1234, 5678])

        Args:
            project_ids: One or more project IDs
        """
        return self.client.paginate(get_projects_by_id, Project, project_id=project_ids, **params)

    @copy_doc_signature(docs._projects_params)
    def search(self, **params) -> Paginator[Project]:
        """Search projects

        .. rubric:: Notes

        * API reference: :v1:`GET /projects <Projects/get_projects>`

        Example:
            Search for projects about invasive species within 400km of Vancouver, BC:

            >>> client.projects.search(
            >>>     q='invasive',
            >>>     lat=49.27,
            >>>     lng=-123.08,
            >>>     radius=400,
            >>>     order_by='distance',
            >>> )
        """
        return self.client.paginate(get_projects, Project, **params)

    def add_observations(
        self, project_id: int, observation_ids: MultiInt, **params
    ) -> ListResponse:
        """Add an observation to a project

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * API reference: :v1:`POST projects/{id}/add <Projects/post_projects_id_add>`
        * API reference: :v1:`POST /project_observations <Project_Observations/post_project_observations>`

        Example:
            >>> client.projects.add_observations(24237, 1234)

        Args:
            project_id: ID of project to add onto
            observation_ids: One or more observation IDs to add
        """
        responses = []
        for observation_id in ensure_list(observation_ids):
            response = self.client.request(
                add_project_observation,
                project_id=project_id,
                observation_id=observation_id,
                auth=True,
                **params,
            )
            responses.append(response)
        return responses

    @copy_doc_signature(docs._project_update_params)
    def update(self, project_id: IntOrStr, **params) -> Project:
        """Update a project

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * Undocumented endpoint; may be subject to braking changes in the future
        * ``admin_attributes`` and ``project_observation_rules_attributes`` each accept a list of dicts
          in the formats shown below. These can be obtained from :py:func:`get_projects`, modified, and
          then passed to this function::

            {
                "admin_attributes": [
                    {"id": int, "role": str, "user_id": int, "_destroy": bool},
                ],
                "project_observation_rules_attributes": [
                    {"operator": str, "operand_type": str, "operand_id": int, "id": int, "_destroy": bool},
                ],
            }

        Example:

            >>> client.projects.update(
            ...     'api-test-project',
            ...     title='Test Project',
            ...     description='This is a test project',
            ...     prefers_rule_native=True,
            ...     access_token=access_token,
            ... )
        """
        response = self.client.request(update_project, project_id, auth=True, **params)
        return Project.from_json(response)

    def add_users(self, project_id: IntOrStr, user_ids: MultiInt, **params) -> Project:
        """Add users to project observation rules

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * This only affects observation rules, **not** project membership

        Example:
            >>> client.projects.add_users(1234, [1234, 5678])

        Args:
            project_id: Either numeric project ID or URL slug
            user_ids: One or more user IDs to add. Only accepts numeric IDs.
        """
        response = self.client.request(add_project_users, project_id, user_ids, auth=True, **params)
        return Project.from_json(response)

    def delete_users(self, project_id: IntOrStr, user_ids: MultiInt, **params) -> Project:
        """Remove users from project observation rules

        .. rubric:: Notes

        * :fa:`lock` :ref:`Requires authentication <auth>`
        * This only affects observation rules, **not** project membership

        Example:
            >>> client.projects.delete_users(1234, [1234, 5678])

        Args:
            project_id: Either numeric project ID or URL slug
            user_ids: One or more user IDs to remove. Only accepts numeric IDs.
        """
        response = self.client.request(
            delete_project_users, project_id, user_ids, auth=True, **params
        )
        return Project.from_json(response)



from logging import getLogger
from os import getenv
from typing import Dict, Optional

from keyring import get_password, set_password
from keyring.errors import KeyringError
from requests import HTTPError, Response

from pyinaturalist.constants import API_V0, API_V1, KEYRING_KEY
from pyinaturalist.exceptions import AuthenticationError
from pyinaturalist.session import ClientSession, get_local_session

logger = getLogger(__name__)


def get_access_token(
    username: Optional[str] = None,
    password: Optional[str] = None,
    app_id: Optional[str] = None,
    app_secret: Optional[str] = None,
    jwt: bool = True,
    refresh: bool = False,
) -> str:
    """Get an access token using the user's iNaturalist username and password, using the
    Resource Owner Password Credentials Flow. Requires registering an iNaturalist app.

    .. rubric:: Notes

    * API reference: https://www.inaturalist.org/pages/api+reference#auth
    * See :ref:`auth` for additional options for storing credentials.
    * This can be used to get either a JWT or OAuth token. These can be used interchangeably for
      many endpoints. JWT is preferred for newer endpoints.

    Examples:

        With direct keyword arguments:

        >>> from pyinaturalist import get_access_token
        >>> access_token = get_access_token(
        >>>     username='my_inaturalist_username',
        >>>     password='my_inaturalist_password',
        >>>     app_id='33f27dc63bdf27f4ca6cd95dd9dcd5df',
        >>>     app_secret='bbce628be722bfe2abd5fc566ba83de4',
        >>> )

        With environment variables or keyring configured:

        >>> access_token = get_access_token()

        If you would like to run custom requests for endpoints not yet implemented in pyinaturalist,
        you can authenticate these requests by putting the token in your HTTP headers as follows:

        >>> import requests
        >>> requests.get(
        >>>     'https://www.inaturalist.org/observations/1234',
        >>>      headers={'Authorization': f'Bearer {access_token}'},
        >>> )

    Args:
        username: iNaturalist username (same as the one you use to login on inaturalist.org)
        password: iNaturalist password (same as the one you use to login on inaturalist.org)
        app_id: OAuth2 application ID
        app_secret: OAuth2 application secret
        jwt: Return a JSON Web Token; otherwise return an OAuth2 access token.
        refresh: Do not use any cached tokens, even if they are not expired

    Raises:
        :py:exc:`requests.HTTPError`: (401) if credentials are invalid
    """
    # First check if we have a previously cached JWT
    session = get_local_session()
    response = _get_jwt(session, only_if_cached=True)
    if response.ok and not refresh:
        logger.info('Using cached access token')
        return response.json()['api_token']

    # Otherwise check for credentials in either args or environment variables
    payload = {
        'username': username or getenv('INAT_USERNAME'),
        'password': password or getenv('INAT_PASSWORD'),
        'client_id': app_id or getenv('INAT_APP_ID'),
        'client_secret': app_secret or getenv('INAT_APP_SECRET'),
        'grant_type': 'password',
    }

    # If any fields were missing, then check the keyring
    if not all(payload.values()):
        payload.update(get_keyring_credentials())
        if all(payload.values()):
            logger.info('Retrieved credentials from keyring')
        else:
            raise AuthenticationError(
                'Not all authentication parameters were provided', response=response
            )

    # Get OAuth access token
    response = session.post(f'{API_V0}/oauth/token', json=payload)
    response.raise_for_status()
    access_token = response.json()['access_token']

    # If specified, use OAuth token to get (and cache) a JWT
    if jwt:
        response = _get_jwt(session, access_token)
        response.raise_for_status()
        access_token = response.json()['api_token']
    return access_token


def validate_token(access_token: str) -> bool:
    """Determine if an access token is valid"""
    session = get_local_session()
    try:
        session.request('GET', f'{API_V1}/users/me', access_token=access_token)
        return True
    except HTTPError:
        return False


def get_keyring_credentials() -> Dict[str, Optional[str]]:
    """Attempt to get iNaturalist credentials from the system keyring

    Returns:
        OAuth-compatible credentials dict
    """
    try:
        return {
            'username': get_password(KEYRING_KEY, 'username'),
            'password': get_password(KEYRING_KEY, 'password'),
            'client_id': get_password(KEYRING_KEY, 'app_id'),
            'client_secret': get_password(KEYRING_KEY, 'app_secret'),
        }
    except KeyringError as e:
        logger.warning(e)
        return {}


def set_keyring_credentials(
    username: str,
    password: str,
    app_id: str,
    app_secret: str,
):
    """
    Store iNaturalist credentials in the system keyring for future use.

    Args:
        username: iNaturalist username
        password: iNaturalist password
        app_id: iNaturalist application ID
        app_secret: iNaturalist application secret
    """
    set_password(KEYRING_KEY, 'username', username)
    set_password(KEYRING_KEY, 'password', password)
    set_password(KEYRING_KEY, 'app_id', app_id)
    set_password(KEYRING_KEY, 'app_secret', app_secret)


def _get_jwt(
    session: ClientSession, access_token: str = '', only_if_cached: bool = False
) -> Response:
    return session.get(
        f'{API_V0}/users/api_token',
        headers={'Authorization': f'Bearer {access_token}'},
        only_if_cached=only_if_cached,  # If True, will return a 504 if not cached
        raise_for_status=False,  # type: ignore
    )



from typing import Optional

from pyinaturalist.constants import IntOrStr, MultiIntOrStr
from pyinaturalist.controllers import BaseController
from pyinaturalist.converters import ensure_list
from pyinaturalist.models import Place
from pyinaturalist.paginator import AutocompletePaginator, Paginator
from pyinaturalist.v1 import get_places_autocomplete, get_places_by_id, get_places_nearby


class PlaceController(BaseController):
    """:fa:`location-dot` Controller for Place requests"""

    def __call__(self, place_id: IntOrStr, **kwargs) -> Optional[Place]:
        """Get a single place by ID

        Example:
            >>> client.places(67591)

        Args:
            place_ids: A single place ID
        """
        return self.from_ids(place_id, **kwargs).one()

    def from_ids(self, place_ids: MultiIntOrStr, **params) -> Paginator[Place]:
        """Get places by ID

        .. rubric:: Notes

        * API reference: :v1:`GET /places/{id} <Places/get_places_id>`

        Example:
            >>> client.places.from_ids([67591, 89191])

        Args:
            place_ids: One or more place IDs
        """
        return self.client.paginate(
            get_places_by_id, Place, place_id=ensure_list(place_ids), **params
        )

    def autocomplete(self, q: Optional[str] = None, **params) -> Paginator[Place]:
        """Given a query string, get places with names starting with the search term

        .. rubric:: Notes

        * API reference: :v1:`GET /places/autocomplete <Places/get_places_autocomplete>`

        Example:
            >>> client.places.autocomplete('Irkutsk')

        Args:
            q: Search query
        """
        return self.client.paginate(
            get_places_autocomplete,
            Place,
            cls=AutocompletePaginator,
            q=q,
            per_page=20,
            **params,
        )

    def nearby(
        self,
        nelat: float,
        nelng: float,
        swlat: float,
        swlng: float,
        name: Optional[str] = None,
        **params,
    ) -> Paginator[Place]:
        """Search for places near a given location

        .. rubric:: Notes

        * API reference: :v1:`GET /places/nearby <get_places_nearby>`

        Example:
            >>> bounding_box = (150.0, -50.0, -149.999, -49.999)
            >>> client.places.nearby(*bounding_box)

        Args:
            nelat: NE latitude of bounding box
            nelng: NE longitude of bounding box
            swlat: SW latitude of bounding box
            swlng: SW longitude of bounding box
            name: Name must match this value
        """
        return self.client.paginate(
            get_places_nearby,
            Place,
            nelat=nelat,
            nelng=nelng,
            swlat=swlat,
            swlng=swlng,
            name=name,
            **params,
        )



"""Models for additional taxon conservation statuses"""

import re
from logging import getLogger
from typing import List, Optional, Union

from attr import define

from pyinaturalist.constants import DateTime
from pyinaturalist.models import (
    BaseModel,
    LazyProperty,
    ListedTaxon,
    Place,
    User,
    datetime_field,
    define_model,
    define_model_custom_init,
    field,
    upper,
)

logger = getLogger(__name__)


@define
class IUCNStatus:
    id: int = field()
    code: str = field()
    description: str = field()


# IUCN Conservation status IDs and codes from https://www.iucnredlist.org
IUCN_STATUSES = [
    IUCNStatus(0, 'NE', 'not evaluated'),
    IUCNStatus(5, 'DD', 'data deficient'),
    IUCNStatus(10, 'LC', 'least concern'),
    IUCNStatus(20, 'NT', 'near threatened'),
    IUCNStatus(30, 'VU', 'vulnerable'),
    IUCNStatus(40, 'EN', 'endangered'),
    IUCNStatus(50, 'CR', 'critically endangered'),
    IUCNStatus(60, 'EW', 'extinct in the wild'),
    IUCNStatus(70, 'EX', 'extinct'),
]
IUCN_STATUSES_BY_CODE = {s.code: s.description for s in IUCN_STATUSES}
IUCN_STATUSES_BY_ID = {s.id: s.description for s in IUCN_STATUSES}

# https://www.natureserve.org/nsexplorer/about-the-data/statuses/conservation-status-categories
NATURESERVE_STATUS_CODES = {
    'X': 'extinct',
    'H': 'possibly extinct',
    '1': 'critically imperiled',
    '2': 'imperiled',
    '3': 'vulnerable',
    '4': 'apparently secure',
    '5': 'secure',
}
NATURESERVE_PATTERN = re.compile(r'\D+([\dXH])(?:\D+([\dXH]))?')

# https://www.dof.gob.mx/normasOficiales/4254/semarnat/semarnat.htm
NORMA_OFICIAL_059_STATUS_CODES = {
    'P': 'en peligro de extinción',
    'A': 'amenazada',
    'PR': 'sujeta a protección especial',
    'EX': 'probablemente extinta en el medio silvestre',
}

# Generic status codes not associated with a specific authority
GENERIC_STATUS_CODES = {
    'CE': 'critically endangered',
    'SE': 'endangered',
    'FE': 'endangered',
    'LE': 'endangered',
    'E': 'endangered',
    'ST': 'threatened',
    'FT': 'threatened',
    'LT': 'threatened',
    'T': 'threatened',
    'LC': 'least concern',
    'SC': 'special concern',
    'C': 'candidate',
}

ALL_STATUS_CODES = {
    **IUCN_STATUSES_BY_CODE,
    **NATURESERVE_STATUS_CODES,
    **NORMA_OFICIAL_059_STATUS_CODES,
    **GENERIC_STATUS_CODES,
}


@define_model_custom_init
class ConservationStatus(BaseModel):
    """:fa:`exclamation-triangle` The conservation status of a taxon in a given location, based on the schema of:

    * ``Taxon.conservation_status`` from :v1:`GET /taxa <Taxa/get_taxa>`
    * ``Observation.taxon.conservation_statuses`` from :v1:`GET /observations <Observations/get_observations>`
    * ``conservation_status`` from
      :v1:`GET /observation/{id}/taxon_summary <Observations/get_observations_id_taxon_summary>`
    """

    authority: str = field(default=None, doc='Data source for conservation status')
    created_at: DateTime = datetime_field(doc='Date and time the record was created')
    description: str = field(default=None, doc='Description of conservation status')
    geoprivacy: str = field(
        default=None,
        doc='Default geoprivacy level; may be obscured or private for protected species',
    )
    iucn: int = field(default=None, doc='IUCN ID, if applicable')
    source_id: int = field(default=None)
    status: str = field(default=None, converter=upper, doc='Short code for conservation status')
    _derived_status_name: str = field(default=None, repr=False)
    _original_status_name: str = field(default=None, repr=False)
    taxon_id: int = field(default=None, doc='Taxon ID')
    updated_at: DateTime = datetime_field(doc='Date and time the record was last updated')
    url: str = field(default=None, doc='Link to data source with more details')

    # Lazy-loaded nested model objects
    place: property = LazyProperty(
        Place.from_json, type=Place, doc='Location that the conservation status applies to'
    )
    updater: User = LazyProperty(User.from_json, type=User, doc='User that last updated the record')  # type: ignore
    user: User = LazyProperty(User.from_json, type=User, doc='User that created the record')  # type: ignore

    # TODO: Are these needed? They appear to be redundant with `status` and `status_name`
    # iucn_status: str = field(default=None)
    # iucn_status_code: str = field(default=None)

    # Attributes that will only be used during init and then omitted
    temp_attrs = ['status_name', 'place_id', 'updater_id', 'user_id']

    def __init__(
        self,
        status_name: Optional[str] = None,
        place_id: Optional[int] = None,
        updater_id: Optional[int] = None,
        user_id: Optional[int] = None,
        **kwargs,
    ):
        self.__attrs_init__(**kwargs)  # type: ignore
        if status_name:
            self._original_status_name = status_name
        if place_id:
            self.place_id = place_id
        if updater_id:
            self.updater_id = updater_id
        if user_id:
            self.user_id = user_id

    @property
    def status_name(self) -> str:
        """Full name of conservation status.

        Note: ``status_name`` from API results can give inconsistent results, so this is derived
        from the status code and authority instead. See ``_original_status_name`` for the original
        value.
        """
        if not self._derived_status_name:
            try:
                self._derived_status_name = translate_status_code(
                    status=self.status, iucn_id=self.iucn, authority=self.authority
                )
            # If it can't be translated, just use the original value
            except (KeyError, ValueError) as e:
                logger.warning(e)
                self._derived_status_name = self._original_status_name or ''
        return self._derived_status_name

    @property
    def display_name(self) -> str:
        """Get conservation status name, code, and place in a format like:
        _'imperiled (S2S3B) in Nova Scotia, CA'_
        """
        # Avoid dusplication when the status name and code are the same (e.g., 'extinct (EXTINCT)')
        status_code_str = f' ({self.status})' if self.status_name != self.status.lower() else ''
        place_str = f' in {self.place_name}' if self.place_name else ''
        return f'{self.status_name}{status_code_str}{place_str}'

    # Wrapper properties to handle inconsistencies between obs, taxa, and taxon_summary endpoints
    @property
    def place_id(self) -> Optional[int]:
        return self.place.id if self.place else None

    @place_id.setter
    def place_id(self, value: int):
        if not self.place:
            self.place = Place()
        self.place.id = value

    @property
    def updater_id(self) -> Optional[int]:
        return self.updater.id if self.updater else None

    @updater_id.setter
    def updater_id(self, value: int):
        if not self.updater:
            self.updater = User()
        self.updater.id = value

    @property
    def user_id(self) -> Optional[int]:
        return self.user.id if self.user else None

    @user_id.setter
    def user_id(self, value: int):
        if not self.user:
            self.user = User()
        self.user.id = value

    @property
    def place_name(self) -> Optional[str]:
        return (self.place.display_name or self.place.name) if self.place else None

    @property
    def _str_attrs(self) -> List[str]:
        return ['status_name', 'status', 'authority', 'place_name']


@define_model
class TaxonSummary(BaseModel):
    """:fa:`dove` :fa:`list` Information about an observation's taxon, within the context
    of the observation's location. Based on the schema of
    :v1:`GET /observations/{id}/taxon_summary <Observations/get_observations_id_taxon_summary>`
    """

    conservation_status: property = LazyProperty(
        ConservationStatus.from_json,
        type=ConservationStatus,
        doc='Conservation status of the taxon in the observation location',
    )
    listed_taxon: property = LazyProperty(
        ListedTaxon.from_json,
        type=ListedTaxon,
        doc='Details about the taxon on an "original" life list',
    )
    wikipedia_summary: str = field(default=None, doc='Taxon summary from Wikipedia article')

    @property
    def _str_attrs(self) -> List[str]:
        return ['conservation_status', 'listed_taxon']


def translate_status_code(
    status: Union[str, int, None] = None,
    iucn_id: Optional[int] = None,
    authority: Optional[str] = None,
) -> str:
    """Translate a conservation status code from a given authority into a descriptive name"""
    status = str(status or '').upper()
    authority = (authority or '').lower()

    # Check status codes by authority
    if authority.startswith('iucn'):
        return IUCN_STATUSES_BY_CODE[status]
    elif authority == 'natureserve':
        return translate_natureserve_code(status)
    elif authority.startswith('norma'):
        return NORMA_OFICIAL_059_STATUS_CODES[status]
    # For other authorities, check both generic and authority-specific status codes
    elif status in ALL_STATUS_CODES:
        return ALL_STATUS_CODES[status]
    elif iucn_id in IUCN_STATUSES_BY_ID:
        return IUCN_STATUSES_BY_ID[iucn_id]
    # Some other authorities use NaturServe format status codes
    elif NATURESERVE_PATTERN.match(status):
        return translate_natureserve_code(status)
    else:
        raise ValueError(
            f'Could not parse conservation status code: {status} from authority: {authority}'
        )


def translate_natureserve_code(status: str) -> str:
    """Translate a NatureServe status code to a status name.

    Note: Status levels may be in ranges, and prefixed with a regional level, e.g. S2S3
    """
    if status in NATURESERVE_STATUS_CODES:
        return NATURESERVE_STATUS_CODES[status]

    match = NATURESERVE_PATTERN.match(status)
    if not match:
        raise ValueError(f'Could not parse NatureServe status code: {status}')

    levels = [int(lvl) for lvl in match.groups() if lvl]
    # Single level
    if len(levels) == 1:
        level = levels[0]
    # Level range spanning 3 levels; pick average
    elif levels[1] - levels[0] == 2:
        level = levels[0] + 1
    # Level range spanning 2 levels; pick higher
    elif levels[1] - levels[0] == 1:
        level = levels[0]
    else:
        raise ValueError(f'Could not parse NatureServe level range: {status}')

    return NATURESERVE_STATUS_CODES[str(level)]



from logging import getLogger
from typing import Dict, List

from pyinaturalist.constants import API_V2, IntOrStr
from pyinaturalist.controllers import BaseController
from pyinaturalist.models import Annotation, ControlledTerm
from pyinaturalist.session import delete, post
from pyinaturalist.v1 import get_controlled_terms, get_controlled_terms_for_taxon

logger = getLogger(__name__)


class AnnotationController(BaseController):
    """:fa:`tag` Controller for Annotation and ControlledTerm requests"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._term_lookup: Dict[int, ControlledTerm] = {}

    @property
    def term_lookup(self) -> Dict[int, ControlledTerm]:
        """Get a lookup table of controlled term IDs to term objects"""
        if not self._term_lookup:
            self._term_lookup = {term.id: term for term in self.all()}
        return self._term_lookup

    def all(self, **params) -> List[ControlledTerm]:
        """List controlled terms and their possible values

        .. rubric:: Notes

        * API reference: :v1:`GET /controlled_terms <Controlled_Terms/get_controlled_terms>`

        Example:
            >>> terms = client.annotations
            >>> pprint(response[0])
            1: Life Stage
                2: Adult
                3: Teneral
                4: Pupa
            ...
        """
        response = get_controlled_terms(**params)
        return ControlledTerm.from_json_list(response['results'])

    def for_taxon(self, taxon_id: int, **params) -> List[ControlledTerm]:
        """List controlled terms that are valid for the specified taxon.

        .. rubric:: Notes

        * API reference: :v1:`GET /controlled_terms/for_taxon <Controlled_Terms/get_controlled_terms_for_taxon>`

        Example:
            >>> client.annotations.for_taxon(12345)

        Args:
            taxon_id: Taxon ID to get controlled terms for

        Raises:
            :py:exc:`.TaxonNotFound`: If an invalid ``taxon_id`` is specified
        """
        response = get_controlled_terms_for_taxon(taxon_id, **params)
        return ControlledTerm.from_json_list(response['results'])

    def lookup(self, annotations: List[Annotation]) -> List[Annotation]:
        """Fill in missing information for the specified annotations. If only term and value IDs are
        present, this will look up, cache, and add complete controlled term details.

        Args:
            annotations: Observation annotations

        Returns:
            Annotation objects with ``controlled_attribute`` and ``controlled_value`` populated
        """
        for annotation in annotations or []:
            term = self.term_lookup.get(annotation.controlled_attribute.id)
            if term:
                annotation.controlled_attribute = term
                annotation.controlled_value = term.get_value_by_id(annotation.controlled_value.id)
            else:
                logger.warning(
                    f'No controlled attribute found for ID: {annotation.controlled_attribute.id}'
                )
        return annotations

    def create(
        self,
        controlled_attribute_id: int,
        controlled_value_id: int,
        resource_id: IntOrStr,
        resource_type: str = 'Observation',
        **params,
    ) -> Annotation:
        """Create a new annotation on an observation.

        Args:
            controlled_attribute_id: Annotation attribute ID
            controlled_value_id: Annotation value ID
            resource_id: Observation ID or UUID
            resource_type: Resource type, if something other than an observation

        Example:
            Add a 'Plant phenology: Flowering' annotation to an observation (via IDs):

            >>> client.annotations.create(12, 13, 164609837)

        Returns:
            The newly created Annotation object
        """
        response = post(
            f'{API_V2}/annotations',
            controlled_attribute_id=controlled_attribute_id,
            controlled_value_id=controlled_value_id,
            resource_id=resource_id,
            resource_type=resource_type,
            **params,
        )
        return Annotation.from_json(response.json()['results'][0])

    def delete(self, uuid: str, **params):
        """Delete an annotation

        Args:
            uuid: Annotation UUID

        Returns:
            Nothing; success means the item has been deleted
        """
        delete(f'{API_V2}/annotations/{uuid}', **params)



"""Reusable template functions used for API documentation.
Each template function contains all or part of an endpoint's request parameters, type annotations,
and docstrings. These are then applied with :py:func:`.copy_doc_signature`.

This is intended to reduce large amounts of duplicated function signatures and corresponding
docstrings. If you're a contributor looking to add a new endpoint, using these templates isn't
necessary. If the new function(s) have a large number of parameters and/or have many parameters in
common with other functions, consider using a template function.

Note: Since the templates are applied dynamically at import time, this adds a tiny amount of overhead
(about 20 milliseconds as of v0.14) to the import time of the library.
"""

# ruff: noqa: E501
from typing import Any, Dict, List, Optional, Union

from pyinaturalist.constants import (
    AnyDate,
    AnyDateTime,
    DateOrStr,
    IntOrStr,
    MultiFile,
    MultiInt,
    MultiIntOrStr,
    MultiStr,
    ObsFieldValues,
)

# Identifications
# --------------------


def _identification_params(
    current_taxon: Optional[bool] = None,
    own_observation: Optional[bool] = None,
    is_change: Optional[bool] = None,
    taxon_active: Optional[bool] = None,
    observation_taxon_active: Optional[bool] = None,
    id: Optional[MultiInt] = None,
    rank: Optional[MultiStr] = None,
    observation_rank: Optional[MultiStr] = None,
    user_id: Optional[MultiIntOrStr] = None,
    user_login: Optional[MultiStr] = None,
    current: Optional[bool] = None,
    category: Optional[MultiStr] = None,
    place_id: Optional[MultiInt] = None,
    quality_grade: Optional[MultiStr] = None,
    taxon_id: Optional[MultiInt] = None,
    observation_taxon_id: Optional[MultiInt] = None,
    iconic_taxon_id: Optional[MultiInt] = None,
    observation_iconic_taxon_id: Optional[MultiInt] = None,
    lrank: Optional[str] = None,
    hrank: Optional[str] = None,
    observation_lrank: Optional[str] = None,
    observation_hrank: Optional[str] = None,
    without_taxon_id: Optional[MultiInt] = None,
    without_observation_taxon_id: Optional[MultiInt] = None,
    d1: Optional[AnyDate] = None,
    d2: Optional[AnyDate] = None,
    observation_created_d1: Optional[AnyDate] = None,
    observation_created_d2: Optional[AnyDate] = None,
    observed_d1: Optional[AnyDate] = None,
    observed_d2: Optional[AnyDate] = None,
    id_above: Optional[int] = None,
    id_below: Optional[int] = None,
):
    """Args:
    current_taxon: ID's taxon is the same it's observation's taxon
    own_observation: ID was added by the observer
    is_change: ID was created as a results of a taxon change
    taxon_active: ID's taxon is currently an active taxon
    observation_taxon_active: Observation's taxon is currently an active taxon
    id: Identification ID
    rank: ID's taxon must have this rank  # Multiple choice
    observation_rank: Observation's taxon must have this rank  # Multiple choice
    user_id: Identifier must have this user ID
    user_login: Identifier must have this user login
    current: Most recent Identification on a observation by a user
    category: Type of identification
    place_id: Observation must occur in this place
    quality_grade: Observation must have this quality grade
    taxon_id: Identification taxa must match the given taxa or their descendants
    observation_taxon_id: Observation taxa must match the given taxa or their descendants
    iconic_taxon_id: Identification iconic taxon ID
    observation_iconic_taxon_id: Observation iconic taxon ID
    lrank: Identification taxon must have this rank or higher
    hrank: Identification taxon must have this rank or lower
    observation_lrank: Observation taxon must have this rank or higher
    observation_hrank: Observation taxon must have this rank or lower
    without_taxon_id: Exclude Identifications of these taxa and their descendants
    without_observation_taxon_id: Exclude Identifications of observations of these taxa and their descendants
    d1: Must be observed on or after this date
    d2: Must be observed on or before this date
    observation_created_d1: Observation must be created on or after this date
    observation_created_d2: Observation must be created on or before this date
    observed_d1: Observation must be observed on or after this date
    observed_d2: Observation must be observed on or before this date
    id_above: Must have an ID above this value
    id_below: Must have an ID below this value
    """


# Messages
# --------------------


def _message_params(
    page: Optional[int] = None,
    box: Optional[str] = None,
    q: Optional[str] = None,
    user_id: Optional[int] = None,
    threads: bool = False,
):
    """Args:
    page: Page number of the results to return
    box: One of 'inbox', 'sent', or 'any'
    q: Search string for message subject and body
    user_id: Get messages to/from this user
    threads: Group results by ``thread_id``, and only get the latest message per thread.
        Incompatible with ``q`` param.
    """


def _message_id(message_id: int):
    """Args:
    message_id: Get the message with this ID. Multiple IDs are allowed.
    """


# Observations
# --------------------


# Params that are in most observation-related endpoints in all API versions
def _observation_common(
    q: Optional[str] = None,
    d1: Optional[AnyDate] = None,
    d2: Optional[AnyDate] = None,
    day: Optional[MultiInt] = None,
    month: Optional[MultiInt] = None,
    year: Optional[MultiInt] = None,
    license: Optional[MultiStr] = None,
    list_id: Optional[int] = None,
    photo_license: Optional[MultiStr] = None,
    out_of_range: Optional[bool] = None,
    quality_grade: Optional[str] = None,
    id: Optional[MultiInt] = None,
    taxon_id: Optional[MultiInt] = None,
    taxon_name: Optional[MultiStr] = None,
    iconic_taxa: Optional[MultiStr] = None,
    updated_since: Optional[AnyDateTime] = None,
):
    """Args:
    q: Search observation properties
    d1: Must be observed on or after this date
    d2: Must be observed on or before this date
    day: Must be observed within this day of the month
    month: Must be observed within this month
    year: Must be observed within this year
    license: Observation must have this license
    photo_license: Must have at least one photo with this license
    out_of_range: Observations whose taxa are outside their known ranges
    list_id: Taxon must be in the list with this ID
    quality_grade: Must have this quality grade
    id: Must have this observation ID
    taxon_id: Only show observations of these taxa and their descendants
    taxon_name: Taxon must have a scientific or common name matching this string
    iconic_taxa: Taxon must by within this iconic taxon
    updated_since: Must be updated since this time
    """


# Observation params that are only in the v1 API
def _observation_v1(
    acc: Optional[bool] = None,
    captive: Optional[bool] = None,
    endemic: Optional[bool] = None,
    geo: Optional[bool] = None,
    id_please: Optional[bool] = None,
    identified: Optional[bool] = None,
    introduced: Optional[bool] = None,
    mappable: Optional[bool] = None,
    native: Optional[bool] = None,
    pcid: Optional[bool] = None,
    photos: Optional[bool] = None,
    popular: Optional[bool] = None,
    sounds: Optional[bool] = None,
    taxon_is_active: Optional[bool] = None,
    threatened: Optional[bool] = None,
    verifiable: Optional[bool] = None,
    not_id: Optional[MultiInt] = None,
    sound_license: Optional[MultiStr] = None,
    observation_fields: Optional[Union[List, Dict]] = None,
    ofv_datatype: Optional[MultiStr] = None,
    place_id: Optional[MultiInt] = None,
    project_id: Optional[MultiInt] = None,
    rank: Optional[MultiStr] = None,
    site_id: Optional[MultiStr] = None,
    without_taxon_id: Optional[MultiInt] = None,
    user_id: Optional[MultiIntOrStr] = None,
    user_login: Optional[MultiStr] = None,
    ident_user_id: Optional[MultiIntOrStr] = None,
    term_id: Optional[MultiInt] = None,
    term_value_id: Optional[MultiInt] = None,
    without_term_value_id: Optional[MultiInt] = None,
    acc_above: Optional[str] = None,
    acc_below: Optional[str] = None,
    acc_below_or_unknown: Optional[str] = None,
    created_d1: Optional[AnyDateTime] = None,
    created_d2: Optional[AnyDateTime] = None,
    created_on: Optional[AnyDate] = None,
    observed_on: Optional[AnyDate] = None,
    unobserved_by_user_id: Optional[int] = None,
    apply_project_rules_for: Optional[str] = None,
    cs: Optional[str] = None,
    csa: Optional[str] = None,
    csi: Optional[MultiStr] = None,
    geoprivacy: Optional[MultiStr] = None,
    taxon_geoprivacy: Optional[MultiStr] = None,
    max_rank: Optional[str] = None,
    min_rank: Optional[str] = None,
    hrank: Optional[str] = None,
    lrank: Optional[str] = None,
    id_above: Optional[int] = None,
    id_below: Optional[int] = None,
    identifications: Optional[str] = None,
    lat: Optional[float] = None,
    lng: Optional[float] = None,
    radius: Optional[float] = None,
    not_in_project: Optional[IntOrStr] = None,
    not_matching_project_rules_for: Optional[IntOrStr] = None,
    search_on: Optional[str] = None,
    viewer_id: Optional[int] = None,
    reviewed: Optional[bool] = None,
    locale: Optional[str] = None,
    preferred_place_id: Optional[int] = None,
    ttl: Optional[str] = None,
):
    """Args:
    acc: Whether or not positional accuracy / coordinate uncertainty has been specified
    captive: Captive or cultivated observations
    endemic: Observations whose taxa are endemic to their location
    geo: Observations that are georeferenced
    id_please: Observations with the **deprecated** 'ID, Please!' flag.
        Note that this will return observations, but that this attribute is no longer used.
    identified: Observations that have community identifications
    introduced: Observations whose taxa are introduced in their location
    mappable: Observations that show on map tiles
    native: Observations whose taxa are native to their location
    pcid: Observations identified by the curator of a project. If the ``project_id`` parameter
        is also specified, this will only consider observations identified by curators of the
        specified project(s)
    photos: Observations with photos
    popular: Observations that have been favorited by at least one user
    sounds: Observations with sounds
    taxon_is_active: Observations of active taxon concepts
    threatened: Observations whose taxa are threatened in their location
    verifiable: Observations with a ``quality_grade`` of either ``needs_id`` or ``research``.
        Equivalent to ``quality_grade=needs_id,research``
    not_id: Must not have this ID
    place_id: Must be observed within the place with this ID
    project_id: Must be added to the project this ID or slug
    rank: Taxon must have this rank
    site_id: Must be affiliated with the iNaturalist network website with this ID
    observation_fields: Must have these observation fields (optionally with values)
    ofv_datatype: Must have an observation field value with this datatype
    sound_license: Must have at least one sound with this license
    without_taxon_id: Exclude observations of these taxa and their descendants
    user_id: Observer must have this user ID or login
    user_login: Observer must have this user login
    ident_user_id: Identifier must have this user ID or login
    term_id: Must have an annotation using this controlled term ID
    term_value_id: Must have an annotation using this controlled value ID.
        Must be combined with the ``term_id`` parameter
    without_term_value_id: Exclude observations with annotations using this controlled value ID.
        Must be combined with the ``term_id`` parameter
    acc_above: Must have an positional accuracy above this value (meters)
    acc_below: Must have an positional accuracy below this value (meters)
    acc_below_or_unknown: Must have an positional accuracy below this value (meters) or unknown
    created_d1: Must be created at or after this time
    created_d2: Must be created at or before this time
    created_on: Must be created on this date
    observed_on: Must be observed on this date
    unobserved_by_user_id: Must not be of a taxon previously observed by this user
    apply_project_rules_for: Must match the rules of the project with this ID or slug
    cs: Taxon must have this conservation status code. If the ``place_id`` parameter is also
        specified, this will only consider statuses specific to that place
    csa: Taxon must have a conservation status from this authority. If the ``place_id`` parameter is
        also specified, this will only consider statuses specific to that place
    csi: Taxon must have this IUCN conservation status. If the ``place_id`` parameter is also
        specified, this will only consider statuses specific to that place
    geoprivacy: Must have this geoprivacy setting
    taxon_geoprivacy: Filter observations by the most conservative geoprivacy applied by a
        conservation status associated with one of the taxa proposed in the current
        identifications.
    hrank: Taxon must have this rank or lower
    lrank: Taxon must have this rank or higher
    id_above: Must have an ID above this value
    id_below: Must have an ID below this value
    identifications: Identifications must meet these criteria
    lat: Must be within a ``radius`` kilometer circle around this lat/lng (lat, lng, radius)
    lng: Must be within a ``radius`` kilometer circle around this lat/lng (lat, lng, radius)
    radius: Must be within a {radius} kilometer circle around this lat/lng (lat, lng, radius)
    not_in_project: Must not be in the project with this ID or slug
    not_matching_project_rules_for: Must not match the rules of the project with this ID or slug
    search_on: Properties to search on, when combined with q. Searches across all properties by
        default
    viewer_id: See reviewed
    reviewed: Observations have been reviewed by the user with ID equal to the value of the
        ``viewer_id`` parameter
    locale: Locale preference for taxon common names
    preferred_place_id: Place preference for regional taxon common names
    ttl: Set the ``Cache-Control`` HTTP header with this value as ``max-age``, in seconds
    """


# Observation params that are only in the v0 API
def _observation_v0(
    has: Optional[MultiStr] = None,
    on: Optional[AnyDate] = None,
    m1: Optional[AnyDate] = None,
    m2: Optional[AnyDate] = None,
    h1: Optional[AnyDate] = None,
    h2: Optional[AnyDate] = None,
    extra: Optional[str] = None,
    response_format: str = 'json',
):
    """Args:
    has: Catch-all for some boolean selectors. This can be used multiple times, e.g.
        ``has=['photos', 'geo']``
    m1: First month of a month range
    m2: Last month of a month range
    h1: First hour of an hour range
    h2: Last hour of an hour range
    on: Filter by date string
    extra: Retrieve additional information.
        **'projects'** returns info about the projects the observations have been added to,
        **'observation_fields'** returns observation field values,
        **'observation_photos'** returns information about the photos' relationship with the
        observation, like their order.
    response_format: A supported response format to return
    """


# Observation params that are only in the v2 API
def _observation_v2(
    fields: Optional[Union[List[str], Dict[str, Any]]] = None,
    except_fields: Optional[List[str]] = None,
):
    """Args:
    fields: Data fields to return in the response
    except_fields: Data fields to exclude from the response (and include all others)
    """


def _observation_histogram(
    date_field: str = 'observed',
    interval: str = 'month_of_year',
):
    """Args:
    date_field: Histogram basis: either when the observation was created or observed
    interval: Time interval for histogram, with groups starting on or contained within the group value.
    """


def _ofvs(observation_id: int, observation_field_id: int, value: Any):
    """Args:
    observation_id: ID of the observation receiving this observation field value
    observation_field_id: ID of the observation field for this observation field value
    value: Value for the observation field
    """


def _create_observation(
    species_guess: Optional[str] = None,
    taxon_id: Optional[int] = None,
    observed_on: Optional[AnyDate] = None,
    observed_on_string: Optional[AnyDate] = None,
    time_zone: Optional[str] = None,
    description: Optional[str] = None,
    tag_list: Optional[MultiStr] = None,
    place_guess: Optional[str] = None,
    latitude: Optional[float] = None,
    longitude: Optional[float] = None,
    map_scale: Optional[int] = None,
    positional_accuracy: Optional[int] = None,
    geoprivacy: Optional[str] = None,
    observation_fields: Optional[ObsFieldValues] = None,
    observation_field_values_attributes: Optional[ObsFieldValues] = None,
    flickr_photos: Optional[MultiInt] = None,
    picasa_photos: Optional[MultiStr] = None,
    facebook_photos: Optional[MultiStr] = None,
    local_photos: Optional[MultiFile] = None,
    photos: Optional[MultiFile] = None,
    sounds: Optional[MultiFile] = None,
    photo_ids: Optional[MultiIntOrStr] = None,
):
    """Args:
    species_guess: Equivalent to the 'What did you see?' field on the observation form.
        iNat will try to choose a single taxon based on this, but it may fail if it's ambiguous
    taxon_id: ID of the taxon to associate with this observation
    observed_on: Alias for ``observed_on_string``; accepts :py:class:`~datetime.datetime` objects.
    observed_on_string: Date/time of the observation. Time zone will default to the user's
        time zone if not specified.
    time_zone: Time zone the observation was made in
    description: Observation description
    tag_list: Comma-separated list of tags
    place_guess: Name of the place where the observation was recorded.
        **Note:** iNat will *not* try to automatically look up coordinates based on this string
    latitude: Latitude of the observation; presumed datum is **WGS84**
    longitude: Longitude of the observation; presumed datum is **WGS84**
    map_scale: Google Maps zoom level (from **0 to 19**) at which to show this observation's map marker.
    positional_accuracy: Positional accuracy of the observation coordinates, in meters
    geoprivacy: Geoprivacy for the observation
    observation_fields: Dict of observation fields in the format ``{id: value}``.
        Alias for ``observation_field_values_attributes``.
    flickr_photos: Flickr photo ID(s) to add as photos for this observation. User must have
        their Flickr and iNat accounts connected, and the user must own the photo(s) on Flickr.
    picasa_photos: Picasa photo ID(s) to add as photos for this observation. User must have
        their Picasa and iNat accounts connected, and the user must own the photo(s) on Picasa.
    facebook_photos: Facebook photo IDs to add as photos for this observation. User must have
        their Facebook and iNat accounts connected, and the user must own the photo on Facebook.
    photos: One or more image files, file-like objects, file paths, or URLs
    sounds: One or more sound files, file-like objects, file paths, or URLs
    photo_ids: One or more IDs of previously uploaded photos to attach to the observation
    """


def _update_observation(
    # _method: str = None,  # Exposed as a client-specific workaround; not needed w/ `requests`
    ignore_photos: bool = True,
):
    """Args:
    ignore_photos: If photos exist on the observation but are missing in the request, ignore them instead of deleting the missing observation photos
    """


# Posts
# --------------------


def _get_posts(
    login: Optional[str] = None,
    project_id: Optional[int] = None,
    page: Optional[int] = None,
    per_page: Optional[int] = None,
):
    """Args:
    login: Return posts by this user
    project_id: Return posts from this project
    page: Pagination page number
    per_page: Number of results to return in a page. The maximum value is generally 200 unless otherwise noted
    """


# Projects
# --------------------


# Note: 'page' parameter is not in API docs, but it appears to work. 'order' does not, however.
def _projects_params(
    q: Optional[str] = None,
    id: Optional[MultiInt] = None,
    not_id: Optional[MultiInt] = None,
    lat: Optional[float] = None,
    lng: Optional[float] = None,
    radius: int = 500,
    featured: Optional[bool] = None,
    noteworthy: Optional[bool] = None,
    place_id: Optional[MultiInt] = None,
    site_id: Optional[int] = None,
    rule_details: Optional[bool] = None,
    type: Optional[MultiStr] = None,
    member_id: Optional[int] = None,
    has_params: Optional[bool] = None,
    has_posts: Optional[bool] = None,
):
    """Args:
    q: Project name must begin with this value
    id: Must have this ID
    not_id: Must not have this ID
    lat: Must be within a ``radius`` kilometer circle around this lat/lng
    lng: Must be within a ``radius`` kilometer circle around this lat/lng
    radius: Distance from center (``(lat, lng)``) to search, in kilometers. Defaults to 500km.
    featured:  Must be marked featured for the relevant site
    noteworthy: Must be marked noteworthy for the relevant site
    place_id: Must be in the place with this ID
    site_id: Site ID that applies to ``featured`` and ``noteworthy``.
        Defaults to the site of the authenticated user, or to the main iNaturalist site
    rule_details: Return more information about project rules, for example return a full taxon
        object instead of simply an ID
    type: Projects must be of this type
    member_id: Project must have member with this user ID
    has_params: Must have search parameter requirements
    has_posts: Must have posts
    order_by: Sort order.
        ``distance`` only applies if lat and lng are specified.
        ``featured`` only applies if ``featured`` or ``noteworthy`` are true.
    """


def _project_observation_params(access_token: str, project_id: int, observation_id: int):
    """Args:
    access_token: An access token required for user authentication, as returned by :py:func:`.get_access_token()`
    project_id: ID of project to add onto
    observation_id: ID of observation to add
    """


def _project_update_params(
    project_id: IntOrStr,
    cover: Optional[str] = None,
    description: Optional[str] = None,
    icon: Optional[str] = None,
    preferred_banner_color: Optional[str] = None,
    prefers_banner_contain: Optional[bool] = None,
    prefers_hide_title: Optional[bool] = None,
    prefers_hide_umbrella_map_flags: Optional[bool] = None,
    prefers_rule_d1: Optional[DateOrStr] = None,
    prefers_rule_d2: Optional[DateOrStr] = None,
    prefers_rule_introduced: Optional[str] = None,
    prefers_rule_members_only: Optional[str] = None,
    prefers_rule_month: Optional[MultiStr] = None,
    prefers_rule_native: Optional[bool] = None,
    prefers_rule_observed_on: Optional[DateOrStr] = None,
    prefers_rule_photos: Optional[bool] = None,
    prefers_rule_quality_grade: Optional[MultiStr] = None,
    prefers_rule_sounds: Optional[bool] = None,
    prefers_rule_term_id: Optional[str] = None,
    prefers_rule_term_value_id: Optional[str] = None,
    prefers_user_trust: Optional[bool] = None,
    project_type: Optional[str] = None,
    title: Optional[str] = None,
    user_id: Optional[int] = None,
    admin_attributes: Optional[List[Dict]] = None,
    project_observation_rules_attributes: Optional[List[Dict]] = None,
):
    """Args:
    project_id: Numeric project ID or slug (the short name shown in project URL)
    cover: Banner image for project page; ideally 760x320 px
    description: Description shown on project page
    icon: Image used as project icon. Should be at least 72x72 px and will be cropped to a square.
    preferred_banner_color: Background color for project banner, as a RGB hex value (e.g., ``'#74ac00'``)
    prefers_banner_contain: Fit banner image without cropping
    prefers_hide_title:
    prefers_hide_umbrella_map_flags:
    prefers_rule_d1: Observation date range to include (start)
    prefers_rule_d2: Observation date range to include (end)
    prefers_rule_observed_on: Exact observation date to include
    prefers_rule_introduced: Only include observations of introduced species
    prefers_rule_native: Only include observations of native species
    prefers_rule_members_only: Only include observations of project members
    prefers_rule_month: Only include observations from these months
    prefers_rule_photos: Only include observations with photos
    prefers_rule_sounds: Only include observations with sounds
    prefers_rule_quality_grade: Only include observations with these quality grades
    prefers_rule_term_id: Only include observations with this annotation (controlled term ID)
    prefers_rule_term_value_id: Only include observations with this annotation value (controlled term value ID)
    prefers_user_trust: Only include observations from trusted users
    project_type: Project type ('umbrella' or 'collection')
    title: Project title
    user_id: User ID of project owner
    admin_attributes: Admin users and their roles
    project_observation_rules_attributes: Rules for observations to include in the project
    """


# Main Search
# --------------------


def _search_params(
    q: str,
    sources: Optional[MultiStr] = None,  # enum
    place_id: Optional[MultiInt] = None,
    locale: Optional[str] = None,
    preferred_place_id: Optional[int] = None,
):
    """Args:
    q: Search query
    sources: Object types to search
    place_id: Results must be associated with this place
    locale: Locale preference for taxon common names
    preferred_place_id: Place preference for regional taxon common names
    """


# Taxa
# --------------------


def _taxon_params(
    q: Optional[str] = None,
    is_active: Optional[bool] = None,
    taxon_id: Optional[int] = None,
    rank: Optional[str] = None,
    max_rank: Optional[str] = None,
    min_rank: Optional[str] = None,
    rank_level: Optional[int] = None,
    locale: Optional[str] = None,
    preferred_place_id: Optional[int] = None,
    all_names: Optional[bool] = None,
):
    """Args:
    q: Name must begin with this value
    is_active: Taxon is active
    taxon_id: Only show taxa with this ID, or its descendants
    rank: Taxon must have this exact rank
    min_rank: Taxon must have this rank or higher; overrides ``rank``
    max_rank: Taxon must have this rank or lower; overrides ``rank``
    rank_level: Taxon must have this rank level. Some example values are 70 (kingdom), 60 (phylum),
        50 (class), 40 (order), 30 (family), 20 (genus), 10 (species), 5 (subspecies)
    locale: Locale preference for taxon common names
    preferred_place_id: Place preference for regional taxon common names
    all_names: Include all taxon names in the response
    """


def _taxon_id_params(
    id_above: Optional[int] = None,
    id_below: Optional[int] = None,
    only_id: Optional[bool] = None,
    parent_id: Optional[int] = None,
):
    """Args:
    id_above: Must have an ID above this value
    id_below: Must have an ID below this value
    only_id: Return only the record IDs
    parent_id: Taxon's parent must have this ID
    """


# Individual/common params
# ------------------------


def _access_token(access_token: Optional[str] = None):
    """Args:
    access_token: An access token for user authentication, as returned by :py:func:`.get_access_token()`
    """


def _bounding_box(
    nelat: Optional[float] = None,
    nelng: Optional[float] = None,
    swlat: Optional[float] = None,
    swlng: Optional[float] = None,
):
    """Args:
    nelat: NE latitude of bounding box
    nelng: NE longitude of bounding box
    swlat: SW latitude of bounding box
    swlng: SW longitude of bounding box
    """


def _geojson_properties(properties: Optional[List[str]] = None):
    """Args:
    properties: Properties from observation results to include as GeoJSON properties
    """


def _name(name: Optional[str] = None):
    """Args:
    name: Name must match this value
    """


def _only_id(only_id: Optional[bool] = None):
    """Args:
    only_id: Return only the record IDs
    """


def _observation_id(observation_id: int):
    """Args:
    observation_id: iNaturalist observation ID
    """


def _project_id(project_id: Optional[int] = None):
    """Args:
    project_id: Only show users who are members of this project
    """


def _pagination(
    page: Optional[int] = None,
    per_page: Optional[int] = None,
    order: Optional[str] = None,
    order_by: Optional[str] = None,
    count_only: Optional[bool] = None,
    reverse: Optional[bool] = None,
):
    """Args:
    page: Page number of results to return
    per_page: Number of results to return in a page. The maximum value is generally 200,
        unless otherwise noted
    order: Sort order
    order_by: Field to sort on
    count_only: Only return a count of results; alias for ``per_page=0``
    reverse: Reverse the order of results; alias for ``order='descending'``
    """


_get_observations = [
    _observation_common,
    _observation_v1,
    _bounding_box,
]


def _search_query(q: Optional[str] = None):
    """Args:
    q: Search query
    """



from datetime import datetime
from typing import Dict, List, Optional

from pyinaturalist.constants import (
    INAT_BASE_URL,
    PROJECT_TYPES,
    Coordinates,
    DateTime,
    JsonResponse,
    TableRow,
)
from pyinaturalist.models import (
    BaseModel,
    LazyProperty,
    ObservationField,
    User,
    coordinate_pair,
    datetime_field,
    datetime_now_field,
    define_model,
    field,
)


@define_model
class ProjectObservation(BaseModel):
    """:fa:`binoculars` Metadata about an observation that has been added to a project"""

    preferences: Dict = field(factory=dict)  # Example: {'allows_curator_coordinate_access': True}
    project: Dict = field(factory=dict)  # Example: {'id': 24237}
    user_id: int = field(default=None)
    uuid: str = field(default=None, doc='Universally unique identifier')
    user: property = LazyProperty(
        User.from_json, type=User, doc='User that added the observation to the project'
    )

    @property
    def _str_attrs(self) -> List[str]:
        return ['project', 'user_id']


@define_model
class ProjectObservationField(ObservationField):
    """:fa:`tag` An :py:class:`.ObservationField` with additional project-specific information"""

    project_observation_field_id: int = field(default=None)
    position: int = field(default=None)
    required: bool = field(default=None)

    @classmethod
    def from_json(cls, value: JsonResponse, **kwargs) -> 'ProjectObservationField':
        """Flatten out nested values"""
        obs_field = value['observation_field']
        obs_field['project_observation_field_id'] = value['id']
        obs_field['position'] = value['position']
        obs_field['required'] = value['required']
        return super(ProjectObservationField, cls).from_json(obs_field, **kwargs)

    @property
    def _str_attrs(self) -> List[str]:
        return ['project_observation_field_id', 'required']


@define_model
class ProjectUser(User):
    """:fa:`user` A :py:class:`.User` with additional project-specific information"""

    project_id: int = field(default=None)
    project_user_id: int = field(default=None)
    role: str = field(default=None)

    @classmethod
    def from_json(cls, value: JsonResponse, **kwargs) -> 'ProjectUser':
        """Flatten out nested values"""
        user = value.get('user', {})
        user['project_id'] = value['project_id']
        user['project_user_id'] = value['id']
        user['role'] = value['role']
        return super(ProjectUser, cls).from_json(user, **kwargs)

    @property
    def _str_attrs(self) -> List[str]:
        return ['project_id', 'project_user_id', 'role']


@define_model
class Project(BaseModel):
    """:fa:`users` An iNaturalist project, based on the schema of
    `GET /projects <https://api.inaturalist.org/v1/docs/#!/Projects/get_projects>`_.
    """

    banner_color: str = field(default=None)
    created_at: datetime = datetime_now_field(doc='Date and time the project was created')
    description: str = field(default=None, doc='Project description')
    header_image_url: str = field(default=None)
    hide_title: bool = field(default=None)
    icon: str = field(default=None, doc='URL for project icon')
    is_umbrella: bool = field(
        default=None,
        doc='Indicates if this is an umbrella project (containing observations from other projects)',
    )
    last_post_at: Optional[datetime] = datetime_field(
        doc='Date and time of the last project journal post'
    )
    location: Coordinates = coordinate_pair()
    observation_requirements_updated_at: Optional[datetime] = datetime_field(
        doc='Date and time of last update for observation requirements'
    )
    place_id: int = field(default=None, doc='Project place ID')
    prefers_user_trust: bool = field(
        default=None,
        doc='Indicates if the project wants users to share hidden coordinates with the project admins',
    )
    project_observation_rules: List[Dict] = field(factory=list, doc='Project observation rules')
    project_type: str = field(default=None, options=PROJECT_TYPES, doc='Project type')  # Enum
    rule_preferences: List[Dict] = field(factory=list)
    search_parameters: List[Dict] = field(factory=list, doc='Filters for observations to include')
    site_features: List[Dict] = field(
        factory=list, doc='Details about if/when the project was featured on inaturalist.org'
    )
    slug: str = field(default=None, doc='URL slug')
    terms: str = field(default=None, doc='Project terms')
    title: str = field(default=None, doc='Project title')
    updated_at: DateTime = datetime_field(doc='Date and time the project was last updated')
    user_ids: List[int] = field(factory=list)

    # Lazy-loaded model objects
    admins: property = LazyProperty(
        ProjectUser.from_json_list, type=List[User], doc='Project admin users'
    )
    project_observation_fields: property = LazyProperty(
        ProjectObservationField.from_json_list,
        type=List[ProjectObservationField],
        doc='Observation fields used by the project',
    )
    user: property = LazyProperty(User.from_json, type=User, doc='User that created the project')

    # Unused attributes
    # flags: List = field(factory=list)
    # header_image_contain: bool = field(default=None)
    # header_image_file_name: str = field(default=None)
    # icon_file_name: str = field(default=None)
    # latitude: float = field(default=None)
    # longitude: float = field(default=None)
    # user_id: int = field(default=None)

    # Aliases
    @property
    def obs_fields(self):
        return self.project_observation_fields

    @property
    def obs_rules(self):
        return self.project_observation_rules

    @property
    def url(self) -> str:
        """Info URL on iNaturalist.org"""
        return f'{INAT_BASE_URL}/projects/{self.id}'

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Title': self.title,
            'Type': self.project_type,
            'URL': self.url,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'title']



from datetime import date, datetime
from logging import getLogger
from typing import List, Union

from pyinaturalist.constants import TableRow
from pyinaturalist.converters import safe_split, try_date, try_datetime, try_int_or_float
from pyinaturalist.models import (
    BaseModel,
    LazyProperty,
    Taxon,
    User,
    datetime_now_field,
    define_model,
    field,
)

logger = getLogger(__name__)
# Mappings from observation field value datatypes to python types/conversion functions
OFV_DATATYPES = {
    'dna': str,
    'date': try_date,
    'datetime': try_datetime,
    'numeric': try_int_or_float,
    'taxon': int,
    'text': str,
    'time': str,
}
OFVValue = Union[date, datetime, float, int, str]


@define_model
class ObservationField(BaseModel):
    """:fa:`tag` An observation field **definition**, based on the schema of
    `GET /observation_fields <https://www.inaturalist.org/pages/api+reference#get-observation_fields>`_.
    """

    allowed_values: List[str] = field(converter=safe_split, factory=list)
    created_at: datetime = datetime_now_field(doc='Date and time the observation field was created')
    datatype: str = field(default=None)  # Enum
    description: str = field(default=None)
    name: str = field(default=None)
    updated_at: datetime = datetime_now_field(
        doc='Date and time the observation field was last updated'
    )
    user_id: int = field(default=None)
    users_count: int = field(default=None)
    uuid: str = field(default=None)
    values_count: int = field(default=None)

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Type': self.datatype,
            'Name': self.name,
            'Description': self.description,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'datatype', 'name', 'description']


@define_model
class ObservationFieldValue(BaseModel):
    """:fa:`tag` An observation field **value**, based on the schema of ``Observation.ofvs``
    from `GET /observations <https://api.inaturalist.org/v1/docs/#!/Observations/get_observations>`_.
    """

    datatype: str = field(default=None)  # Enum
    field_id: int = field(default=None)
    name: str = field(default=None)
    taxon_id: int = field(default=None)
    user_id: int = field(default=None)
    uuid: str = field(default=None)
    value: OFVValue = field(default=None)
    taxon: property = LazyProperty(
        Taxon.from_json, type=Taxon, doc='Taxon that the observation field applies to'
    )
    user: property = LazyProperty(
        User.from_json, type=User, doc='User that applied the observation field value'
    )

    # Unused attrbiutes
    # name_ci: str = field(default=None)
    # value_ci: int = field(default=None)

    # Convert value by datatype
    def __attrs_post_init__(self):
        if self.datatype in OFV_DATATYPES and self.value is not None:
            converter = OFV_DATATYPES[self.datatype]
            try:
                self.value = converter(self.value)
            except (TypeError, ValueError) as e:
                logger.debug(f'Failed to convert {self.name}="{self.value}" to {converter}: {e}')
                self.value = None

    @property
    def _row(self) -> TableRow:
        return {
            'ID': self.id,
            'Type': self.datatype,
            'Name': self.name,
            'Value': self.value,
        }

    @property
    def _str_attrs(self) -> List[str]:
        return ['id', 'datatype', 'name', 'value']



"""Placeholder module for backwards-compatibility"""

# ruff: noqa: F401, F403, F405
from warnings import warn

from pyinaturalist.auth import get_access_token
from pyinaturalist.constants import ListResponse
from pyinaturalist.paginator import paginate_all
from pyinaturalist.v0 import *

msg = (
    'The module `pyinaturalist.rest_api` is deprecated; please use `from pyinaturalist import ...`'
)
warn(DeprecationWarning(msg), stacklevel=2)


def get_all_observation_fields(**params) -> ListResponse:
    """:fas:`triangle-exclamation` Deprecated; use ``get_observation_fields(page='all')`` instead"""
    msg = "get_all_observation_fields() is deprecated; please use get_observation_fields(page='all') instead"
    warn(DeprecationWarning(msg), stacklevel=2)
    return paginate_all(get_observation_fields, method='page', **params)['results']


def add_photo_to_observation(*args, **kwargs):
    return upload_photos(*args, **kwargs)



# ruff: noqa: F401
from datetime import date, datetime, timedelta
from os.path import abspath, dirname, join
from pathlib import Path
from typing import IO, TYPE_CHECKING, Any, BinaryIO, Dict, Iterable, List, Optional, Tuple, Union

from dateutil.relativedelta import relativedelta
from platformdirs import user_data_dir

# iNaturalist URLs
API_V0 = 'https://www.inaturalist.org'
API_V1 = 'https://api.inaturalist.org/v1'
API_V2 = 'https://api.inaturalist.org/v2'
EXPORT_URL = 'https://www.inaturalist.org/observations/export'
GBIF_TAXON_BASE_URL = 'https://www.gbif.org/species'
INAT_BASE_URL = API_V0
INAT_REPO = 'https://raw.githubusercontent.com/inaturalist/inaturalist/main'
ICONIC_TAXA_BASE_URL = f'{INAT_REPO}/app/assets/images/iconic_taxa'
PHOTO_BASE_URL = 'https://static.inaturalist.org/photos'
PHOTO_CC_BASE_URL = 'https://inaturalist-open-data.s3.amazonaws.com/photos'
PHOTO_INFO_BASE_URL = 'https://www.inaturalist.org/photos'

# Prefix used for keyring entries
KEYRING_KEY = '/inaturalist'

# Pagination settings
PER_PAGE_RESULTS = 200  # Default number of records per page for paginated queries
LARGE_REQUEST_WARNING = 5000  # Show a warning for queries that will return over this many results

# Rate-limiting and retry settings
CONNECT_TIMEOUT = 5
MAX_DELAY = 60  # Maximum time to wait for rate-limiting before aborting
MAX_FILESIZE = 20000000  # 20MB maximum file size for uploads
REQUEST_BURST_RATE = 5
REQUESTS_PER_SECOND = 1
REQUESTS_PER_MINUTE = 60
REQUESTS_PER_DAY = 10000
REQUEST_TIMEOUT = 20
REQUEST_RETRIES = 5  # Maximum number of retries for a failed request
RETRY_BACKOFF = 0.5  # Exponential backoff factor for retries
RETRY_STATUSES = (500, 502, 503, 504)

# HTTP methods that apply to write-only dry-run mode
WRITE_HTTP_METHODS = ['PATCH', 'POST', 'PUT', 'DELETE']

# Project directories
PROJECT_DIR = abspath(dirname(dirname(__file__)))
DATA_DIR = Path(user_data_dir()) / 'pyinaturalist'
DOCS_DIR = join(PROJECT_DIR, 'docs')
DOWNLOAD_DIR = join(PROJECT_DIR, 'downloads')
EXAMPLES_DIR = join(PROJECT_DIR, 'examples')
SAMPLE_DATA_DIR = join(PROJECT_DIR, 'test', 'sample_data')

# Cache settings
CACHE_EXPIRATION = {
    'api.inaturalist.org/*autocomplete': timedelta(days=30),
    'api.inaturalist.org/v*/controlled_terms*': timedelta(days=7),
    'api.inaturalist.org/v*/places*': timedelta(days=7),
    'api.inaturalist.org/v*/taxa*': timedelta(days=7),
    'www.inaturalist.org/users/api_token': timedelta(days=1),
    f'{PHOTO_CC_BASE_URL}/*': -1,
    f'{PHOTO_BASE_URL}/*': -1,
    f'{ICONIC_TAXA_BASE_URL}/*': -1,
    '*': timedelta(minutes=30),
}
CACHE_FILE = join(DATA_DIR, 'api_requests.db')
RATELIMIT_FILE = join(DATA_DIR, 'api_ratelimit.db')
DEFAULT_LOCK_PATH = join(DATA_DIR, 'api_ratelimit.lock')

# Response formats supported by v0 GET /observations endpoint
OBSERVATION_FORMATS = ['atom', 'csv', 'dwc', 'json', 'kml', 'widget']

# Taxon ID and name of main taxa 'categories' that can be filtered on
ICONIC_TAXA = {
    0: 'Unknown',
    1: 'Animalia',
    3: 'Aves',
    20978: 'Amphibia',
    26036: 'Reptilia',
    40151: 'Mammalia',
    47178: 'Actinopterygii',
    47115: 'Mollusca',
    47119: 'Arachnida',
    47158: 'Insecta',
    47126: 'Plantae',
    47170: 'Fungi',
    48222: 'Chromista',
    47686: 'Protozoa',
}
ICONIC_EMOJI = {
    0: '❓',
    1: '🐾',
    3: '🐦',
    20978: '🐸',
    26036: '🦎',
    40151: '😺',
    47178: '🐠',
    47115: '🐌',
    47119: '🕷️',
    47158: '🦋',
    47126: '🌿',
    47170: '🍄',
    48222: '🟢',
    47686: '🦠',
}
ROOT_TAXON_ID = 48460

# Taxonomic ranks that can be filtered on, and numeric values for comparison
# Source: https://github.com/inaturalist/inaturalist/blob/main/app/models/taxon.rb
UNRANKED = 90  # Used for comparing unranked taxa
RANK_LEVELS = {
    'infrahybrid': 5,
    'form': 5,
    'variety': 5,
    'subspecies': 5,
    'hybrid': 10,
    'species': 10,
    'complex': 11,
    'subsection': 12,
    'section': 13,
    'subgenus': 15,
    'genushybrid': 20,
    'genus': 20,
    'subtribe': 24,
    'tribe': 25,
    'supertribe': 26,
    'subfamily': 27,
    'family': 30,
    'epifamily': 32,
    'superfamily': 33,
    'zoosubsection': 33.5,
    'zoosection': 34,
    'parvorder': 34.5,
    'infraorder': 35,
    'suborder': 37,
    'order': 40,
    'superorder': 43,
    'subterclass': 44,
    'infraclass': 45,
    'subclass': 47,
    'class': 50,
    'superclass': 53,
    'subphylum': 57,
    'phylum': 60,
    'kingdom': 70,
    'unranked': UNRANKED,
    'stateofmatter': 100,
}
RANKS = list(RANK_LEVELS.keys())[:-2]

# The main 7 taxonomic ranks, used for condensed display
COMMON_RANKS = [
    'species',
    'genus',
    'family',
    'order',
    'class',
    'phylum',
    'kingdom',
]

# Additional equivalents that iNat accepts; see request_params.normalize_rank() for more variations
RANK_EQUIVALENTS = {
    'division': 'phylum',
    'gen': 'genus',
    'sp': 'species',
    'spp': 'species',
    'infraspecies': 'subspecies',
    'ssp': 'subspecies',
    'subsp': 'subspecies',
    'trinomial': 'subspecies',
    'var': 'variety',
}


# Options for multiple choice parameters (non-endpoint-specific)
CC_LICENSES = ['CC-BY', 'CC-BY-NC', 'CC-BY-ND', 'CC-BY-SA', 'CC-BY-NC-ND', 'CC-BY-NC-SA', 'CC0']
ALL_LICENSES = CC_LICENSES + ['ALL RIGHTS RESERVED']
COMMUNITY_ID_STATUSES = ['most_agree', 'most_disagree', 'some_agree']
CONSERVATION_STATUSES = ['LC', 'NT', 'VU', 'EN', 'CR', 'EW', 'EX']
ESTABLISTMENT_MEANS = ['introduced', 'native', 'endemic']
EXTRA_PROPERTIES = ['fields', 'identifications', 'projects']
GEOPRIVACY_LEVELS = ['obscured', 'obscured_private', 'open', 'private']
HAS_PROPERTIES = ['photo', 'geo']
HISTOGRAM_DATE_FIELDS = ['created', 'observed']
HISTOGRAM_INTERVALS = ['year', 'month', 'week', 'day', 'hour', 'month_of_year', 'week_of_year']
ICON_SIZES = {'icon': 32, 'square': 75, 'small': 75, 'medium': 200, 'large': 200, 'original': 200}
ID_CATEGORIES = ['improving', 'supporting', 'leading', 'maverick']
INBOXES = ['inbox', 'sent', 'any']
ORDER_DIRECTIONS = ['asc', 'desc']
PHOTO_SIZES = ['square', 'small', 'medium', 'large', 'original']
PLACE_CATEGORIES = ['standard', 'community']
PROJECT_TYPES = ['assessment', 'bioblitz', 'collection', 'umbrella']
QUALITY_GRADES = ['casual', 'needs_id', 'research']
SEARCH_PROPERTIES = ['names', 'tags', 'description', 'place']
SOURCES = ['places', 'projects', 'taxa', 'users']

# Endpoint-specific options for multiple choice parameters
V0_OBS_ORDER_BY_PROPERTIES = ['date_added', 'observed_on']
V1_OBS_ORDER_BY_PROPERTIES = ['created_at', 'id', 'observed_on', 'species_guess', 'votes']
V2_OBS_ORDER_BY_PROPERTIES = [
    'created_at',
    'id',
    'observed_on',
    'species_guess',
    'updated_at',
    'votes',
    'random',
]
PROJECT_ORDER_BY_PROPERTIES = ['created', 'distance', 'featured', 'recent_posts', 'updated']

# Multiple-choice request parameters, with keys mapped to their possible choices (non-endpoint-specific)
# TODO: python 3.11 supports Literal[*list] syntax, which would be useful here
MULTIPLE_CHOICE_PARAMS = {
    'box': INBOXES,
    'category': ID_CATEGORIES,
    'csi': CONSERVATION_STATUSES,
    'date_field': HISTOGRAM_DATE_FIELDS,
    'extra': EXTRA_PROPERTIES,
    'geoprivacy': GEOPRIVACY_LEVELS,
    'has': HAS_PROPERTIES,
    'hrank': RANKS,
    'iconic_taxa': list(ICONIC_TAXA.values()),
    'identifications': COMMUNITY_ID_STATUSES,
    'interval': HISTOGRAM_INTERVALS,
    'license': CC_LICENSES,
    'lrank': RANKS,
    'max_rank': RANKS,
    'min_rank': RANKS,
    'observation_hrank': RANKS,
    'observation_lrank': RANKS,
    'observation_rank': RANKS,
    'order': ORDER_DIRECTIONS,
    'photo_license': CC_LICENSES,
    'quality_grade': QUALITY_GRADES,
    'rank': RANKS,
    'search_on': SEARCH_PROPERTIES,
    'sound_license': CC_LICENSES,
    'sources': SOURCES,
    'taxon_geoprivacy': GEOPRIVACY_LEVELS,
    'type': PROJECT_TYPES,
}

# Request parameters from all API versions that accept date or datetime strings
DATETIME_PARAMS = [
    'created_after',
    'created_d1',
    'created_d2',
    'created_on',
    'd1',
    'd2',
    'newer_than',
    'observation_created_d1',
    'observation_created_d2',
    'observed_d1',
    'observed_d2',
    'observed_on',
    'observed_on_string',
    'older_than',
    'on',
    'prefers_rule_d1',
    'prefers_rule_d2',
    'since',
    'updated_since',
]
DATETIME_SHORT_FORMAT = '%b %d, %Y'

# Type aliases
Coordinates = Tuple[float, float]
AnyDate = Union[date, datetime, str]
AnyDateTime = Union[datetime, str]
AnyFile = Union[IO, Path, str]
DateTime = datetime
DateOrInt = Union[date, datetime, int]
DateOrStr = Union[date, datetime, str]
DateOrDatetime = Union[date, datetime]
DateRange = Tuple[DateOrDatetime, DateOrDatetime]
Dimensions = Tuple[int, int]
FileOrPath = Union[BinaryIO, Path, str]
GeoJson = Dict[str, Any]
HistogramResponse = Dict[DateOrInt, int]
IntOrStr = Union[int, str]
JsonResponse = Dict[str, Any]
ListResponse = List[Dict[str, Any]]
ObsFieldValues = Union[Dict, List[Dict]]
RequestParams = Dict[str, Any]
ResponseResult = Dict[str, Any]
ResponseOrResults = Union[JsonResponse, Iterable[ResponseResult]]
ResponseOrFile = Union[AnyFile, JsonResponse]
MultiFile = Union[FileOrPath, Iterable[FileOrPath]]
MultiInt = Union[int, Iterable[int]]
MultiStr = Union[str, Iterable[str]]
MultiIntOrStr = Union[MultiInt, MultiStr]
TableRow = Dict[str, Any]
TemplateFunction = Any  # Cannot use Callable/Protocol, as these will not allow a mix of signatures
TimeInterval = Union[str, timedelta, relativedelta]

# For type checking purposes, some nullable attrs need to be marked as Optional.
# For documentation purposes, this is redundant since all keyword args are optional.
if TYPE_CHECKING:
    Coordinates = Optional[Coordinates]  # type: ignore
    DateTime = Optional[DateTime]  # type: ignore



"""Helper functions for processing and validating request parameters.
The main purpose of these functions is to support some python-specific conveniences and translate
them into standard API request parameters, along with client-side request validation.

Also see :py:mod:`pyinaturalist.converters` for type conversion utilities not specific to request
params.
"""

import re
from datetime import date, datetime, timedelta
from inspect import signature
from logging import getLogger
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Union

from dateutil.parser import parse as parse_date
from dateutil.relativedelta import relativedelta

from pyinaturalist.constants import *  # noqa: F401, F403  # Imports for backwards-compatibility
from pyinaturalist.constants import (
    DATETIME_PARAMS,
    MULTIPLE_CHOICE_PARAMS,
    RANK_EQUIVALENTS,
    RANKS,
    DateOrStr,
    DateRange,
    MultiInt,
    RequestParams,
    TimeInterval,
)
from pyinaturalist.converters import convert_isoformat, ensure_list

# Common parameters that can be passed to all API functions, and notes on where they are used
COMMON_PARAMS = [
    'access_token',  # Used in session.prepare_request()
    'allow_str_ids',  # Used in session.prepare_request()
    'dry_run',  # Used in session.request()
    'expire_after',  # Passed to requests_cache.CachedSession.send()
    'limit',  # Used in paginator.Paginator
    'session',  # Used in session.request()
    'timeout',  # Used in session.ClientSession.send()
    'user_agent',  # Used in session.prepare_request()
]

# Time interval options used by observation histogram
INTERVALS: Dict[str, Union[timedelta, relativedelta]] = {
    'hour': timedelta(hours=1),
    'day': timedelta(days=1),
    'week': relativedelta(weeks=1),
    'month': relativedelta(months=1),
    'year': relativedelta(years=1),
}

RANK_PARAMS = [
    'rank',
    'lrank',
    'hrank',
    'min_rank',
    'max_rank',
    'observation_rank',
    'observation_lrank',
    'observation_hrank',
]

MULTIPLE_CHOICE_ERROR_MSG = (
    'Parameter "{}" must have one of the following values: {}\n\tValue provided: {}'
)

MAX_URL_LENGTH = 2048  # Depends on server and browser. This is is on the lower end.

logger = getLogger(__name__)


def preprocess_request_body(body: Optional[RequestParams]) -> Optional[RequestParams]:
    """Perform type conversions, sanity checks, etc. on JSON-formatted request body"""
    if not body:
        return None
    for resource in ['project', 'observation']:
        if resource in body:
            body[resource] = preprocess_request_params(body[resource], convert_lists=False)
    else:
        body = preprocess_request_params(body, convert_lists=False)
    return body


def preprocess_request_params(
    params: Optional[RequestParams], convert_lists: bool = True
) -> RequestParams:
    """Perform type conversions, sanity checks, etc. on request parameters"""
    if not params:
        return {}

    params = normalize_rank_params(params)
    params = validate_multiple_choice_params(params)
    params = convert_pagination_params(params)
    params = convert_bool_params(params)
    params = convert_datetime_params(params)
    params = convert_observation_field_filters(params)
    if convert_lists:
        params = convert_list_params(params)
    params = strip_empty_values(params)
    params, _ = split_common_params(params)
    return params


def convert_bool_params(params: RequestParams) -> RequestParams:
    """Convert any boolean request parameters to javascript-style boolean strings"""
    for k, v in params.items():
        if isinstance(v, bool):
            params[k] = str(v).lower()
    return params


def convert_url_ids(url: str, ids: Optional[MultiInt] = None, allow_str_ids: bool = False) -> str:
    """If one or more resources are requested by ID, validate and update the request URL
    accordingly"""
    if not ids:
        return url
    if not allow_str_ids:
        ids = _validate_ids(ids)
    return url.rstrip('/') + '/' + _join_list(ids)


def _join_list(obj: Any) -> str:
    """Join a list of items into an API-compatible (comma-delimited) string"""
    if isinstance(obj, list) or isinstance(obj, tuple):
        return ','.join(map(str, obj))
    elif obj is None:
        return ''
    else:
        return str(obj)


def _validate_ids(ids: Any) -> List[int]:
    """Ensure ID(s) are all valid, and convert to a comma-delimited string if there are multiple

    Raises:
        :py:exc:`ValueError` if any values are not valid integers
    """
    try:
        return [int(value) for value in ensure_list(ids, split_str_list=True)]
    except (TypeError, ValueError) as e:
        raise ValueError(f'Invalid ID(s): {ids}; must specify integers only') from e


def convert_datetime_params(params: RequestParams) -> RequestParams:
    """Convert any dates, datetimes, or timestamps in other formats into ISO 8601 strings.

    API behavior note: params that take date but not time info will accept a full timestamp and
    just ignore the time, so it's safe to parse both date and datetime strings into timestamps

    Raises:
        :py:exc:`dateutil.parser._parser.ParserError` if a date/datetime format is invalid
    """
    for k, v in params.items():
        if isinstance(v, (date, datetime)) or (isinstance(v, str) and k in DATETIME_PARAMS):
            params[k] = convert_isoformat(v)
    return params


def convert_list_params(params: RequestParams) -> RequestParams:
    """Convert any list parameters into an API-compatible (comma-delimited) string.
    Will be url-encoded by requests. For example: `['k1', 'k2', 'k3'] -> k1%2Ck2%2Ck3`
    """
    return {k: _join_list(v) for k, v in params.items()}


def convert_observation_params(params):
    """Some common parameter conversions needed by observation CRUD endpoints"""
    params = convert_observation_field_params(params)
    if params.get('observed_on'):
        params['observed_on_string'] = params.pop('observed_on')

    # Split out photos and sounds to upload separately
    photos = ensure_list(params.pop('local_photos', None))
    photos.extend(ensure_list(params.pop('photos', None)))  # Alias for 'local_photos'
    sounds = ensure_list(params.pop('sounds', None))
    photo_ids = ensure_list(params.pop('photo_ids', None))

    # Split API request params from common function args
    params, kwargs = split_common_params(params)

    # ignore_photos must be 1 rather than true; 0 does not work, so just remove if false
    if params.pop('ignore_photos', True):
        kwargs['ignore_photos'] = 1

    return photos, sounds, photo_ids, params, kwargs


def convert_observation_field_params(params: RequestParams) -> RequestParams:
    """Translate simplified format of observation field values into API-compatible format"""
    if 'observation_fields' in params:
        params['observation_field_values_attributes'] = params.pop('observation_fields')
    obs_fields = params.get('observation_field_values_attributes')
    if isinstance(obs_fields, dict):
        params['observation_field_values_attributes'] = [
            {'observation_field_id': k, 'value': v} for k, v in obs_fields.items()
        ]
    return params


def convert_observation_field_filters(params: RequestParams) -> RequestParams:
    """Convert observation field filters from simplified format to full request params"""
    fields = params.pop('observation_fields', None)
    if not fields:
        return params
    # List of fields to filter on (any value)
    elif isinstance(fields, (list, tuple)):
        for field in fields:
            params[f'field:{field}'] = ''
    # List of fields to filter on (specific values)
    elif isinstance(fields, Mapping):
        for field, value in fields.items():
            params[f'field:{field}'] = value or ''
    return params


def convert_pagination_params(params: RequestParams) -> RequestParams:
    """Allow ``count_only=True`` as a slightly more intuitive shortcut to only get a count of
    results"""
    if params.pop('count_only', False) is True:
        params['per_page'] = 0
    if params.pop('reverse', False) is True:
        params['order'] = 'descending'
    if params.get('page') is not None and params['page'] < 1:
        raise ValueError('Page number must be a positive integer')
    return params


def convert_rank_range(params: RequestParams) -> RequestParams:
    """If min and/or max rank is specified in params, translate into a list of ranks"""

    def _get_rank_index(rank: str) -> int:
        rank = normalize_rank(rank)
        if rank not in RANKS:
            raise ValueError(f'Invalid rank: {rank}')
        return RANKS.index(rank)

    min_rank, max_rank = params.pop('min_rank', None), params.pop('max_rank', None)
    if min_rank or max_rank:
        # Use indices in RANKS list to determine range of ranks to include
        min_rank_index = _get_rank_index(min_rank) if min_rank else 0
        max_rank_index = _get_rank_index(max_rank) + 1 if max_rank else len(RANKS)
        params['rank'] = RANKS[min_rank_index:max_rank_index]
    return params


def normalize_rank(rank: str) -> str:
    """Normalize and validate a taxonomic rank, alias, or abbreviation"""
    rank = re.sub(r'[-_\.\s]', '', rank).lower()

    # Check if it's an alias/equivalent rank that iNaturalist accepts
    if rank in RANK_EQUIVALENTS:
        rank = RANK_EQUIVALENTS[rank]
    # Check if it's a unique prefix of a rank
    elif rank not in RANKS:
        matching_ranks = [r for r in RANKS if r.startswith(rank)]
        rank = matching_ranks[0] if len(matching_ranks) == 1 else rank

    return rank


def normalize_rank_params(params: RequestParams) -> RequestParams:
    """Normalize any taxonomic ranks in request params"""
    for k in RANK_PARAMS:
        if isinstance(params.get(k), (tuple, list)):
            params[k] = [normalize_rank(r) for r in params[k]]
        elif params.get(k):
            params[k] = normalize_rank(params[k])
    return params


def get_interval_ranges(
    start: DateOrStr, end: DateOrStr, interval: TimeInterval
) -> List[DateRange]:
    """Given a date range and a time interval, split the range into a list of smaller ranges

    Args:
        start: Start date or string (inclusive)
        end: End date or string (inclusive)
        interval: Time interval (delta or alias: 'hour', 'day', 'month', or 'year')

    Returns:
        List of date ranges of size ``interval``, in the format: ``[(start_date, end_date), ...]``
    """
    if isinstance(interval, str):
        interval = INTERVALS[interval]
    if isinstance(start, str):
        start = parse_date(start)
    if isinstance(end, str):
        end = parse_date(end)

    ranges = []
    while start <= end:
        # API date/datetime request params are inclusive, so subtract 1 minute from end date
        ranges.append((start, start + interval - timedelta(minutes=1)))
        start += interval
    return ranges


def get_valid_kwargs(func: Callable, kwargs: Dict) -> Dict:
    """Get the subset of non-None ``kwargs`` that are valid params for ``func``"""
    sig_params = list(signature(func).parameters)
    return {k: v for k, v in kwargs.items() if k in sig_params and v is not None}


def split_common_params(params: RequestParams) -> Tuple[RequestParams, RequestParams]:
    """Split out common keyword args (for pyinaturalist functions) from request params (for API)"""
    kwargs = {k: params.pop(k, None) for k in COMMON_PARAMS}
    return params, kwargs


def strip_empty_values(values: Dict) -> Dict:
    """Remove any dict items with empty or ``None`` values.
    Observation field filters are an exception (e.g. ``field:foo=``).
    """
    return {k: v for k, v in values.items() if v or v in [False, 0, 0.0] or k.startswith('field:')}


def validate_multiple_choice_params(params: RequestParams) -> RequestParams:
    """Verify that multiple-choice request parameters contain a valid value.

    **Note:** This does not check endpoint-specific params, i.e., those that have the same name
    but different values across different endpoints.

    Returns:
        Parameters with modifications (if any)

    Raises:
        :py:exc:`ValueError`;
            Error message will contain info on all validation errors, if there are multiple
    """
    # Collect info on any validation errors
    errors = []
    for key, choices in MULTIPLE_CHOICE_PARAMS.items():
        params, error_msg = _validate_multiple_choice_param(params, key, choices)
        if error_msg:
            errors.append(error_msg)

    # Combine all messages (if multiple) into one error message
    if errors:
        raise ValueError('\n'.join(errors))
    return params


def validate_multiple_choice_param(
    params: RequestParams, key: str, choices: Iterable
) -> RequestParams:
    """Verify that a multiple-choice request parameter contains valid value(s);
    if not, raise an error.
    **Used for endpoint-specific params.**

    Returns:
        Parameters with modifications (if any)

    Raises:
        :py:exc:`ValueError`
    """
    params, error_msg = _validate_multiple_choice_param(params, key, choices)
    if error_msg:
        raise ValueError(error_msg)
    return params


def _validate_multiple_choice_param(
    params: RequestParams, key: str, choices: Iterable
) -> Tuple[RequestParams, Optional[str]]:
    """Verify that a multiple-choice request parameter contains valid value(s);
    if not, return an error message.

    Returns:
        Parameters with modifications (if any), and a validation error message (if any)
    """

    def is_valid(value, choices):
        if not value:
            return True
        if not isinstance(value, list):
            value = [value]
        return all(v in choices for v in value)

    def normalize(value):
        if not value:
            return value
        if isinstance(value, list):
            return [v.replace(' ', '_') for v in value]
        return value.replace(' ', '_')

    error_msg = None
    if key in params:
        params[key] = normalize(params[key])
    if not is_valid(params.get(key), choices):
        error_msg = MULTIPLE_CHOICE_ERROR_MSG.format(key, choices, params[key])
    return params, error_msg



"""Data models that represent iNaturalist API response objects.
See :ref:`data-models` section for usage details.
"""

# ruff: noqa: F401, E402
# isort: skip_file
from datetime import datetime, timezone
from typing import Callable, Dict, Iterable, Optional, Union

import attr
from attr import define, validators

from pyinaturalist.constants import JsonResponse, ResponseOrResults
from pyinaturalist.converters import convert_lat_long, try_datetime
from pyinaturalist.models.base import BaseModel, BaseModelCollection, T, load_json
from pyinaturalist.models.lazy_property import LazyProperty, add_lazy_attrs, get_lazy_properties


# Aliases and minor helper functions used by model classes
# --------------------------------------------------------

# Attrs class decorators with the most commonly used options
define_model: Callable = define(auto_attribs=False, field_transformer=add_lazy_attrs)
define_model_custom_init: Callable = define(
    auto_attribs=False, init=False, field_transformer=add_lazy_attrs
)
define_model_collection: Callable = define(auto_attribs=False, order=False, slots=False)


def field(
    doc: str = '', options: Optional[Iterable] = None, metadata: Optional[Dict] = None, **kwargs
):
    """A field with extra metadata for documentation and options"""
    metadata = metadata or {}
    metadata['doc'] = doc
    metadata['options'] = options
    return attr.field(**kwargs, metadata=metadata)


# TODO: Case-insensitive version
def is_in(options: Iterable):
    """Validator for an optional multiple-choice attribute"""
    return validators.in_(list(options) + [None])


def coordinate_pair(doc: Optional[str] = None, **kwargs):
    """Field containing a pair of coordinates"""
    return field(
        default=None,
        converter=convert_lat_long,
        doc=doc or 'Location in ``(latitude, longitude)`` decimal degrees',
        **kwargs,
    )


def datetime_field(**kwargs):
    """Field that converts date/time strings into datetime objects"""
    return field(default=None, converter=try_datetime, **kwargs)


def datetime_now_field(**kwargs):
    """Field that converts date/time strings into datetime objects, and defaults to the current time"""
    return field(converter=try_datetime, factory=lambda: datetime.now(timezone.utc), **kwargs)


def upper(value) -> Optional[str]:
    """Converter to change a string to uppercase, if applicable"""
    return value.upper() if isinstance(value, str) else None


# Models imported in order of dependencies
# --------------------------------------------------------

from pyinaturalist.models.media import IconPhoto, Photo, Sound
from pyinaturalist.models.place import Place
from pyinaturalist.models.user import User, UserCount, UserCounts
from pyinaturalist.models.checklist import Checklist, EstablishmentMeans, ListedTaxon
from pyinaturalist.models.conservation_status import ConservationStatus, TaxonSummary
from pyinaturalist.models.taxon import Taxon, TaxonCount, TaxonCounts, LifeList, make_tree
from pyinaturalist.models.controlled_term import (
    Annotation,
    ControlledTerm,
    ControlledTermCount,
    ControlledTermCounts,
    ControlledTermValue,
)
from pyinaturalist.models.identification import Comment, Identification
from pyinaturalist.models.message import Message
from pyinaturalist.models.observation_field import ObservationField, ObservationFieldValue
from pyinaturalist.models.project import (
    Project,
    ProjectObservation,
    ProjectObservationField,
    ProjectUser,
)
from pyinaturalist.models.observation import (
    Observation,
    Observations,
    Application,
    Fave,
    Flag,
    QualityMetric,
    Vote,
)
from pyinaturalist.models.search import SearchResult


# Type aliases involving model objects
ModelObjects = Union[BaseModel, BaseModelCollection, Iterable[BaseModel]]
ResponseOrObject = Union[BaseModel, JsonResponse]
ResponseOrObjects = Union[ModelObjects, ResponseOrResults]

# Short aliases for some of the longer class names
ID = Identification
OFD = ObservationField
OFV = ObservationFieldValue



from pyinaturalist.constants import API_V1, ListResponse
from pyinaturalist.converters import convert_all_coordinates, convert_all_timestamps
from pyinaturalist.docs import document_request_params
from pyinaturalist.docs import templates as docs
from pyinaturalist.session import get


@document_request_params(docs._get_posts)
def get_posts(**params) -> ListResponse:
    """Search posts

    .. rubric:: Notes

    * API reference: :v1:`GET /posts <Posts/get_posts>`

    Example:
        Get journal posts from user 'username'

        >>> response = get_posts(login='username')

    Returns:
        List containing journal posts from the iNaturalist site
    """
    response = get(f'{API_V1}/posts', **params)

    posts = response.json()
    posts = convert_all_coordinates(posts)
    posts = convert_all_timestamps(posts)

    return posts
